/*
 * config.c -- configuration of the simulator
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "common.h"
#include "config.h"
#include "pipe.h"


/**************************************************************/


#define errprint0(f)		fprintf(stderr, f)
#define errprint1(f,a)		fprintf(stderr, f, a)
#define errprint2(f,a,b)	fprintf(stderr, f, a, b)
#define errprint3(f,a,b,c)	fprintf(stderr, f, a, b, c)
#define panic(x)		{ x; \
				  errprint0("!\n"); \
				  exit(-1); }

#define BUF_SIZE		100


/**************************************************************/


typedef struct {
  char name[20];	/* symbolic name */
  int *v;		/* internal name */
  int defVal;		/* default value */
  int minVal, maxVal;	/* min and max legal values */
  Bool powerOfTwo;	/* must it be a power of two? */
} PV;


typedef enum {
  assoc, blksz, setsz, gran, vctsz,
  wrb, wra, acctm, citm, cotm, prts
} CacheParam;

typedef struct {
  char name[20];	/* symbolic name */
  CacheParam v;		/* internal code */
  int defVal;		/* default value */
  int minVal, maxVal;	/* min and max legal values */
  Bool powerOfTwo;	/* must it be a power of two? */
} CPV;


typedef struct {
  char name[8];		/* symbolic name */
  InternalOpcode v;	/* internal code */
  int defVal;		/* default value */
} OP;


/**************************************************************/


static FILE *configFile;	/* input comes from here */
static char buffer[BUF_SIZE];	/* input lines go here */
static char token[BUF_SIZE];	/* and tokens are copied to here */
static char *bufPtr = buffer;	/* this is our current position */
static Bool tokenPrescanned;	/* did we scan the next token already? */

static int fetchBufSize;
static int writeBufSize;
static int reorderBufSize;
static int memBusBytes;
static Bool hwPT;
static int maxCycs = 60;

static PV pv[] = {
  { "fetchbuffer",       &fetchBufSize,       4,    1,   INT_MAX, false },
  { "writebuffer",       &writeBufSize,       2,    1,   INT_MAX, false },
  { "reorderbuffer",     &reorderBufSize,     5,    1,   INT_MAX, false },
  { "renameregs",        &maxRenameRegs,      5,    1,   INT_MAX, false },
  { "memslots",          &maxMemSlots,        2,    1,   INT_MAX, false },
  { "localregs",         &lringSize,          256,  256, 1024,    true  },
  { "fetchmax",          &fetchMax,           2,    1,   INT_MAX, false },
  { "dispatchmax",       &dispatchMax,        1,    1,   INT_MAX, false },
  { "peekahead",         &peekAhead,          1,    0,   INT_MAX, false },
  { "commitmax",         &commitMax,          1,    1,   INT_MAX, false },
  { "fremmax",           &fremMax,            1,    1,   INT_MAX, false },
  { "denin",             &deninPenalty,       1,    0,   INT_MAX, false },
  { "denout",            &denoutPenalty,      1,    0,   INT_MAX, false },
  { "writeholdingtime",  &holdingTime,        0,    0,   INT_MAX, false },
  { "memaddresstime",    &memAddrTime,        20,   1,   INT_MAX, false },
  { "memreadtime",       &memReadTime,        20,   1,   INT_MAX, false },
  { "memwritetime",      &memWriteTime,       20,   1,   INT_MAX, false },
  { "membusbytes",       &memBusBytes,        8,    8,   INT_MAX, true  },
  { "branchpredictbits", &bpN,                0,    0,   8,       false },
  { "branchaddressbits", &bpA,                0,    0,   32,      false },
  { "branchhistorybits", &bpB,                0,    0,   32,      false },
  { "branchdualbits",    &bpC,                0,    0,   32,      false },
  { "hardwarepagetable", (int *) &hwPT,       1,    0,   1,       false },
  { "disablesecurity",   (int *) &noSecurity, 0,    0,   1,       false },
  { "memchunksmax",      &memChunksMax,       1000, 1,   INT_MAX, false },
  { "hashprime",         &hashPrime,          2003, 2,   INT_MAX, false }
};

static CPV cpv[] = {
  { "associativity", assoc, 1, 1, INT_MAX, true  },
  { "blocksize",     blksz, 8, 8, 8192,    true  },
  { "setsize",       setsz, 1, 1, INT_MAX, true  },
  { "granularity",   gran,  8, 8, 8192,    true  },
  { "victimsize",    vctsz, 0, 0, INT_MAX, true  },
  { "writeback",     wrb,   0, 0, 1,       false },
  { "writeallocate", wra,   0, 0, 1,       false },
  { "accesstime",    acctm, 1, 1, INT_MAX, false },
  { "copyintime",    citm,  1, 1, INT_MAX, false },
  { "copyouttime",   cotm,  1, 1, INT_MAX, false },
  { "ports",         prts,  1, 1, INT_MAX, false }
};

static OP op[] = {
  { "mul0",  mul0,  10 },
  { "mul1",  mul1,  10 },
  { "mul2",  mul2,  10 },
  { "mul3",  mul3,  10 },
  { "mul4",  mul4,  10 },
  { "mul5",  mul5,  10 },
  { "mul6",  mul6,  10 },
  { "mul7",  mul7,  10 },
  { "mul8",  mul8,  10 },
  { "div",   div,   60 },
  { "sh",    sh,    1  },
  { "mux",   mux,   1  },
  { "sadd",  sadd,  1  },
  { "mor",   mor,   1  },
  { "fadd",  fadd,  4  },
  { "fmul",  fmul,  4  },
  { "fdiv",  fdiv,  40 },
  { "fsqrt", fsqrt, 40 },
  { "fint",  fint,  4  },
  { "fix",   fix,   2  },
  { "flot",  flot,  2  },
  { "feps",  feps,  4  }
};

static int pvSize;
static int cpvSize;
static int opSize;

static InternalOpcode intOp[256] = {
  trap,   fcmp,   funeq,  funeq,  fadd,   fix,    fadd,   fix,
  flot,   flot,   flot,   flot,   flot,   flot,   flot,   flot,
  fmul,   feps,   feps,   feps,   fdiv,   fsqrt,  frem,   fint,
  mul,    mul,    mul,    mul,    div,    div,    div,    div,
  add,    add,    addu,   addu,   sub,    sub,    subu,   subu,
  addu,   addu,   addu,   addu,   addu,   addu,   addu,   addu,
  cmp,    cmp,    cmpu,   cmpu,   sub,    sub,    subu,   subu,
  sh,     sh,     sh,     sh,     sh,     sh,     sh,     sh,
  br,     br,     br,     br,     br,     br,     br,     br,
  br,     br,     br,     br,     br,     br,     br,     br,
  pbr,    pbr,    pbr,    pbr,    pbr,    pbr,    pbr,    pbr,
  pbr,    pbr,    pbr,    pbr,    pbr,    pbr,    pbr,    pbr,
  cset,   cset,   cset,   cset,   cset,   cset,   cset,   cset,
  cset,   cset,   cset,   cset,   cset,   cset,   cset,   cset,
  zset,   zset,   zset,   zset,   zset,   zset,   zset,   zset,
  zset,   zset,   zset,   zset,   zset,   zset,   zset,   zset,
  ld,     ld,     ld,     ld,     ld,     ld,     ld,     ld,
  ld,     ld,     ld,     ld,     ld,     ld,     ld,     ld,
  ld,     ld,     ld,     ld,     ld,     ld,     ld,     ld,
  ld,     ld,     ld,     ld,     prego,  prego,  go,     go,
  st,     st,     st,     st,     st,     st,     st,     st,
  st,     st,     st,     st,     st,     st,     st,     st,
  st,     st,     st,     st,     st,     st,     st,     st,
  st,     st,     st,     st,     st,     st,     pushgo, pushgo,
  or,     or,     orn,    orn,    nor,    nor,    xor,    xor,
  and,    and,    andn,   andn,   nand,   nand,   nxor,   nxor,
  bdif,   bdif,   wdif,   wdif,   tdif,   tdif,   odif,   odif,
  mux,    mux,    sadd,   sadd,   mor,    mor,    mor,    mor,
  set,    set,    set,    set,    addu,   addu,   addu,   addu,
  or,     or,     or,     or,     andn,   andn,   andn,   andn,
  noop,   noop,   pushj,  pushj,  set,    set,    put,    put,
  pop,    resume, save,   unsave, sync,   noop,   get,    trip
};

static int intStages[MAX_REAL_CMD + 1];	/* internal opcode -> stages */
static int stages[256];			/* mmixOpcode -> stages */


/**************************************************************/


static void getToken(void) {
  char *p, *q;

  /* check if token already present */
  if (tokenPrescanned) {
    tokenPrescanned = false;
    return;
  }
  /* scan past white space */
  while (1) {
    if (*bufPtr == '\0' ||
        *bufPtr == '\n' ||
        *bufPtr == '%') {
      if (fgets(buffer, BUF_SIZE, configFile) == NULL) {
        strcpy(token, "end");
        return;
      }
      if (strlen(buffer) == BUF_SIZE - 1 &&
          buffer[BUF_SIZE - 2] != '\n') {
        panic(errprint1("config file line too long: '%s...'", buffer));
      }
      bufPtr = buffer;
    } else
    if (!isspace(*bufPtr)) {
      break;
    } else {
      bufPtr++;
    }
  }
  /* copy token */
  for (p = bufPtr, q = token; !isspace(*p) && *p != '%'; p++, q++) {
    *q = *p;
  }
  bufPtr = p;
  *q = '\0';
}


static int getInt(void) {
  int v;

  getToken();
  if (sscanf(token, "%d", &v) != 1) {
    return -1;
  }
  return v;
}


static Cache *newCache(char *name) {
  Cache *c;

  c = calloc(1, sizeof(Cache));
  if (c == NULL) {
    panic(errprint1("Cannot allocate %s", name));
  }
  c->aa = 1;  /* default associativity, should equal cpv[0].defVal */
  c->bb = 8;  /* default blocksize */
  c->cc = 1;  /* default setsize */
  c->gg = 8;  /* default granularity */
  c->vv = 0;  /* default victimsize */
  c->repl = random;  /* default replacement policy */
  c->vrepl = random;  /* default victim replacement policy */
  c->mode = 0;  /* default mode: write-through and write-around */
  c->accessTime = 1;
  c->copyInTime = 1;
  c->copyOutTime = 1;
  c->filler.ctl = &(c->fillerCtl);
  c->fillerCtl.ptrA = (void *) c;
  c->fillerCtl.go.o.l = 4;
  c->flusher.ctl = &(c->flusherCtl);
  c->flusherCtl.ptrA = (void *) c;
  c->flusherCtl.go.o.l = 4;
  c->ports = 1;
  c->name = name;
  return c;
}


static void ppol(ReplacePolicy *rr) {
  getToken();
  if (strcmp(token, "random") == 0) {
    *rr = random;
  } else
  if (strcmp(token, "serial") == 0) {
    *rr = serial;
  } else
  if (strcmp(token, "pseudolru") == 0) {
    *rr = pseudoLRU;
  } else
  if (strcmp(token, "lru") == 0) {
    *rr = trueLRU;
  } else {
    /* oops, we should rescan that token */
    tokenPrescanned = true;
  }
}


static void pcs(Cache *c) {
  int j, n;

  getToken();
  for (j = 0; j < cpvSize; j++) {
    if (strcmp(token, cpv[j].name) == 0) {
      break;
    }
  }
  if (j == cpvSize) {
    panic(errprint1("Config syntax error: %s is not a cache parameter name",
                    token));
  }
  n = getInt();
  if (n < cpv[j].minVal) {
    panic(errprint2("Config error: %s must be >= %d",
                    cpv[j].name, cpv[j].minVal));
  }
  if (n > cpv[j].maxVal) {
    panic(errprint2("Config error: %s must be <= %d",
                    cpv[j].name, cpv[j].maxVal));
  }
  if (cpv[j].powerOfTwo && (n & (n - 1)) != 0) {
    panic(errprint1("Config error: %s must be a power of 2",
                    cpv[j].name));
  }
  switch (cpv[j].v) {
    case assoc:
      c->aa = n;
      ppol(&(c->repl));
      break;
    case blksz:
      c->bb = n;
      break;
    case setsz:
      c->cc = n;
      break;
    case gran:
      c->gg = n;
      break;
    case vctsz:
      c->vv = n;
      ppol(&(c->vrepl));
      break;
    case wrb:
      c->mode = (c->mode & ~WRITE_BACK) + n * WRITE_BACK;
      break;
    case wra:
      c->mode = (c->mode & ~WRITE_ALLOC) + n * WRITE_ALLOC;
      break;
    case acctm:
      if (n > maxCycs) {
        maxCycs = n;
      }
      c->accessTime = n;
      break;
    case citm:
      if (n > maxCycs) {
        maxCycs = n;
      }
      c->copyInTime = n;
      break;
    case cotm:
      if (n > maxCycs) {
        maxCycs = n;
      }
      c->copyOutTime = n;
      break;
    case prts:
      c->ports = n;
      break;
  }
}


static int lg(int n) {
  int j, l;

  l = 0;
  for (j = n; j != 0; j >>= 1) {
    l++;
  }
  return l - 1;
}


static void allocCache(Cache *c, char *name) {
  int j, k;

  if (c->bb < c->gg) {
    panic(errprint1("Config error: blocksize of %s is less than granularity",
                    name));
  }
  if (name[1] == 'T' && c->bb != 8) {
    panic(errprint1("Config error: blocksize of %s must be 8",
                    name));
  }
  c->a = lg(c->aa);
  c->b = lg(c->bb);
  c->c = lg(c->cc);
  c->g = lg(c->gg);
  c->v = lg(c->vv);
  c->tagMask = -(1 << (c->b + c->c));
  if (c->a + c->b + c->c >= 32) {
    panic(errprint1("Config error: %s has >= 4 gigabytes of data",
                    name));
  }
  if (c->gg != 8 && (c->mode & WRITE_ALLOC) == 0) {
    panic(errprint2("Config error: %s does write-around with granularity %d",
                    name, c->gg));
  }
  /* allocate the cache sets for cache c <32 begin> */
  c->set = calloc(c->cc, sizeof(CacheSet));
  if (c->set == NULL) {
    panic(errprint1("Cannot allocate cache sets for %s",
                    name));
  }
  for (j = 0; j < c->cc; j++) {
    c->set[j] = calloc(c->aa, sizeof(CacheBlock));
    if (c->set[j] == NULL) {
      panic(errprint2("Cannot allocate cache blocks (set %d) of %s",
                      j, name));
    }
    for (k = 0; k < c->aa; k++) {
      c->set[j][k].tag.h = SIGN_BIT;  /* invalid tag */
      c->set[j][k].dirty = calloc(c->bb >> c->g, sizeof(char));
      if (c->set[j][k].dirty == NULL) {
        panic(errprint3("Cannot allocate dirty bits (block %d, set %d) of %s",
                        k, j, name));
      }
      c->set[j][k].data = calloc(c->bb >> 3, sizeof(Octa));
      if (c->set[j][k].data == NULL) {
        panic(errprint3("Cannot allocate data (block %d, set %d) of %s",
                        k, j, name));
      }
    }
  }
  /* allocate the cache sets for cache c <32 end> */
  if (c->vv) {
    /* allocate the victim cache for cache c <33 begin> */
    c->victim = calloc(c->vv, sizeof(CacheBlock));
    if (c->victim == NULL) {
      panic(errprint1("Cannot allocate blocks for victim cache of %s",
                      name));
    }
    for (k = 0; k < c->vv; k++) {
      c->victim[k].tag.h = SIGN_BIT;  /* invalid tag */
      c->victim[k].dirty = calloc(c->bb >> c->g, sizeof(char));
      if (c->victim[k].dirty == NULL) {
        panic(errprint2("Cannot allocate dirty bits (block %d, victim) of %s",
                        k, name));
      }
      c->victim[k].data = calloc(c->bb >> 3, sizeof(Octa));
      if (c->victim[k].data == NULL) {
        panic(errprint2("Cannot allocate data (block %d, victim) of %s",
                        k, name));
      }
    }
    /* allocate the victim cache for cache c <33 end> */
  }
  c->inbuf.dirty = calloc(c->bb >> c->g, sizeof(char));
  if (c->inbuf.dirty == NULL) {
    panic(errprint1("Cannot allocate dirty bits for inbuffer of %s",
                    name));
  }
  c->inbuf.data = calloc(c->bb >> 3, sizeof(Octa));
  if (c->inbuf.data == NULL) {
    panic(errprint1("Cannot allocate data for inbuffer of %s",
                    name));
  }
  c->outbuf.dirty = calloc(c->bb >> c->g, sizeof(char));
  if (c->outbuf.dirty == NULL) {
    panic(errprint1("Cannot allocate dirty bits for outbuffer of %s",
                    name));
  }
  c->outbuf.data = calloc(c->bb >> 3, sizeof(Octa));
  if (c->outbuf.data == NULL) {
    panic(errprint1("Cannot allocate data for outbuffer of %s",
                    name));
  }
  if (name[0] != 'S') {
    /* allocate reader coroutines for cache c <34 begin> */
    c->reader = calloc(c->ports, sizeof(Coroutine));
    if (c->reader == NULL) {
      panic(errprint1("Cannot allocate readers for %s",
                      name));
    }
    for (j = 0; j < c->ports; j++) {
      c->reader[j].stage = VANISH;
      c->reader[j].name =
        (name[0] == 'D' ?
          (name[1] == 'T' ? "DTreader" : "Dreader") :
          (name[1] == 'T' ? "ITreader" : "Ireader"));
    }
    /* allocate reader coroutines for cache c <34 end> */
  }
}


/**************************************************************/


void mmixConfig(char *fileName) {
  int i, j, n;
  Coroutine *p;

  configFile = fopen(fileName, "r");
  if (configFile == NULL) {
    panic(errprint1("Cannot open configuration file %s", fileName));
  }
  /* initialize to defaults <17 begin> */
  pvSize = sizeof(pv) / sizeof(pv[0]);
  cpvSize = sizeof(cpv) / sizeof(cpv[0]);
  opSize = sizeof(op) / sizeof(op[0]);
  ITcache = newCache("ITcache");
  DTcache = newCache("DTcache");
  Icache = NULL;
  Dcache = NULL;
  Scache = NULL;
  for (j = 0; j < pvSize; j++) {
    *(pv[j].v) = pv[j].defVal;
  }
  for (j = 0; j < opSize; j++) {
    pipeSeq[op[j].v][0] = op[j].defVal;
    pipeSeq[op[j].v][1] = 0;  /* one stage */
  }
  /* initialize to defaults <17 end> */
  /* count and allocate the functional units <18 begin> */
  funitCount = 0;
  while (strcmp(token, "end") != 0) {
    getToken();
    if (strcmp(token, "unit") == 0) {
      funitCount++;
      getToken();
      getToken();
    }
  }
  funit = calloc(funitCount + 1, sizeof(Func));
  if (funit == NULL) {
    panic(errprint0("Cannot allocate the functional units"));
  }
  /* the last unit is for TRIP and TRAP instructions */
  strcpy(funit[funitCount].name, "%%");
  funit[funitCount].ops[0] = 0x80000000;  /* TRAP */
  funit[funitCount].ops[7] = 0x00000001;  /* TRIP */
  /* count and allocate the functional units <18 end> */
  /* record all the specs <19 begin> */
  rewind(configFile);
  funitCount = 0;
  token[0] = '\0';
  while (strcmp(token, "end") != 0) {
    getToken();
    if (strcmp(token, "end") == 0) {
      break;
    }
    /* if token is a parameter name, process a PV spec <20 begin> */
    for (j = 0; j < pvSize; j++) {
      if (strcmp(token, pv[j].name) == 0) {
        n = getInt();
        if (n < pv[j].minVal) {
          panic(errprint2("Config error: %s must be >= %d",
                          pv[j].name, pv[j].minVal));
        }
        if (n > pv[j].maxVal) {
          panic(errprint2("Config error: %s must be <= %d",
                          pv[j].name, pv[j].maxVal));
        }
        if (pv[j].powerOfTwo && (n & (n - 1)) != 0) {
          panic(errprint1("Config error: %s must be a power of 2",
                          pv[j].name));
        }
        *(pv[j].v) = n;
        break;
      }
    }
    if (j < pvSize) {
      continue;
    }
    /* if token is a parameter name, process a PV spec <20 end> */
    /* if token is a cache name, process a cache spec <21 begin> */
    if (strcmp(token, "ITcache") == 0) {
      pcs(ITcache);
      continue;
    }
    if (strcmp(token, "DTcache") == 0) {
      pcs(DTcache);
      continue;
    }
    if (strcmp(token, "Icache") == 0) {
      if (Icache == NULL) {
        Icache = newCache("Icache");
      }
      pcs(Icache);
      continue;
    }
    if (strcmp(token, "Dcache") == 0) {
      if (Dcache == NULL) {
        Dcache = newCache("Dcache");
      }
      pcs(Dcache);
      continue;
    }
    if (strcmp(token, "Scache") == 0) {
      if (Icache == NULL) {
        Icache = newCache("Icache");
      }
      if (Dcache == NULL) {
        Dcache = newCache("Dcache");
      }
      if (Scache == NULL) {
        Scache = newCache("Scache");
      }
      pcs(Scache);
      continue;
    }
    /* if token is a cache name, process a cache spec <21 end> */
    /* if token is an operation name, process a pipe spec <24 begin> */
    for (j = 0; j < opSize; j++) {
      if (strcmp(token, op[j].name) == 0) {
        for (i = 0; ; i++) {
          n = getInt();
          if (n < 0) {
            break;
          }
          if (n == 0) {
            panic(errprint0("Config error: pipeline cycles must be positive"));
          }
          if (n > 255) {
            panic(errprint0("Config error: pipeline cycles must be <= 255"));
          }
          if (n > maxCycs) {
            maxCycs = n;
          }
          if (i >= PIPE_LIMIT) {
            panic(errprint1("Config error: more than %d pipeline stages",
                            PIPE_LIMIT));
          }
          pipeSeq[op[j].v][i] = n;
        }
        tokenPrescanned = true;
        break;
      }
    }
    if (j < opSize) {
      continue;
    }
    /* if token is an operation name, process a pipe spec <24 end> */
    if (strcmp(token, "unit") == 0) {
      /* process a functional spec <25 begin> */
      getToken();
      if (strlen(token) > 15) {
        panic(errprint1("Config error: '%s' is more than 15 characters long",
                        token));
      }
      strcpy(funit[funitCount].name, token);
      getToken();
      if (strlen(token) != 64) {
        panic(errprint1("Config error: unit %s does not have 64 digit specs",
                        funit[funitCount].name));
      }
      i = 0;
      n = 0;
      for (j = 0; j < 64; j++) {
        if (token[j] >= '0' && token[j] <= '9') {
          n = (n << 4) + (token[j] - '0');
        } else
        if (token[j] >= 'a' && token[j] <= 'f') {
          n = (n << 4) + (token[j] - 'a' + 10);
        } else
        if (token[j] >= 'A' && token[j] <= 'F') {
          n = (n << 4) + (token[j] - 'A' + 10);
        } else {
          panic(errprint1("Config error: '%c' is not a hex digit",
                          token[j]));
        }
        if ((j & 0x7) == 0x7) {
          funit[funitCount].ops[i] = n;
          i++;
          n = 0;
        }
      }
      funitCount++;
      continue;
      /* process a functional spec <25 end> */
    }
    panic(errprint1("Config syntax error: spec cannot start with '%s'",
                    token));
  }
  /* record all the specs <19 end> */
  /* allocate coroutines in each functional unit <26 begin> */
  /* first build table of pipeline stages needed for each opcode <27 begin> */
  for (j = div; j <= MAX_PIPE_OP; j++) {
    intStages[j] = strlen(pipeSeq[j]);
  }
  for (; j <= MAX_REAL_CMD; j++) {
    intStages[j] = 1;
  }
  n = 0;
  for (j = mul0; j <= mul8; j++) {
    if (strlen(pipeSeq[j]) > n) {
      n = strlen(pipeSeq[j]);
    }
  }
  intStages[mul] = n;
  intStages[ld] = 2;
  intStages[st] = 2;
  intStages[frem] = 2;
  for (j = 0; j < 256; j++) {
    stages[j] = intStages[intOp[j]];
  }
  /* first build table of pipeline stages needed for each opcode <27 end> */
  for (j = 0; j <= funitCount; j++) {
    /* determine the number of stages, n, needed by funit[j] <29 begin> */
    n = 0;
    for (i = 0; i < 256; i++) {
      if (((funit[j].ops[i >> 5] << (i & 0x1F)) & 0x80000000) != 0 &&
          stages[i] > n) {
        n = stages[i];
      }
    }
    if (n == 0) {
      panic(errprint1("Config error: unit %s does not do anything",
                      funit[j].name));
    }
    /* determine the number of stages, n, needed by funit[j] <29 end> */
    funit[j].st = n;
    funit[j].co = calloc(n, sizeof(Coroutine));
    if (funit[j].co == NULL) {
      panic(errprint0("Cannot allocate coroutines"));
    }
    for (i = 0; i < n; i++) {
      funit[j].co[i].name = funit[j].name;
      funit[j].co[i].stage = i + 1;
    }
  }
  /* allocate coroutines in each functional unit <26 end> */
  /* allocate the caches <35 begin> */
  allocCache(ITcache, "ITcache");
  ITcache->filler.name = "ITfiller";
  ITcache->filler.stage = FILL_FROM_VIRT;
  allocCache(DTcache, "DTcache");
  DTcache->filler.name = "DTfiller";
  DTcache->filler.stage = FILL_FROM_VIRT;
  if (Icache) {
    allocCache(Icache, "Icache");
    Icache->filler.name = "Ifiller";
    Icache->filler.stage = FILL_FROM_MEM;
  }
  if (Dcache) {
    allocCache(Dcache, "Dcache");
    Dcache->filler.name = "Dfiller";
    Dcache->filler.stage = FILL_FROM_MEM;
    Dcache->flusher.name = "Dflusher";
    Dcache->flusher.stage = FLUSH_TO_MEM;
  }
  if (Scache) {
    allocCache(Scache, "Scache");
    if (Scache->bb < Icache->bb) {
      panic(errprint0("Config error: Scache blocks < Icache blocks"));
    }
    if (Scache->bb < Dcache->bb) {
      panic(errprint0("Config error: Scache blocks < Dcache blocks"));
    }
    if (Scache->gg != Dcache->gg) {
      panic(errprint0("Config error: Scache ganularity differs from Dcache"));
    }
    Icache->filler.stage = FILL_FROM_S;
    Dcache->filler.stage = FILL_FROM_S;
    Dcache->flusher.stage = FLUSH_TO_S;
    Scache->filler.name = "Sfiller";
    Scache->filler.stage = FILL_FROM_MEM;
    Scache->flusher.name = "Sflusher";
    Scache->flusher.stage = FLUSH_TO_MEM;
  }
  /* allocate the caches <35 end> */
  /* allocate the scheduling queue <36 begin> */
  busWords = memBusBytes >> 3;
  j = (memReadTime < memWriteTime ? memWriteTime : memReadTime);
  n = 1;
  if (Scache != NULL && Scache->bb > n) {
    n = Scache->bb;
  }
  if (Icache != NULL && Icache->bb > n) {
    n = Icache->bb;
  }
  if (Dcache != NULL && Dcache->bb > n) {
    n = Dcache->bb;
  }
  n = memAddrTime + ((int) (n + busWords - 1) / busWords) * j;
  if (n > maxCycs) {
    maxCycs = n;  /* now maxCycs bounds the waiting time */
  }
  ringSize = maxCycs + 1;
  ring = calloc(ringSize, sizeof(Coroutine));
  if (ring == NULL) {
    panic(errprint0("Cannot allocate the scheduling ring"));
  }
  for (p = ring; p < ring + ringSize; p++) {
    p->name = "";  /* header nodes are nameless */
    p->stage = MAX_STAGE;
  }
  /* allocate the scheduling queue <36 end> */
  /* touch up last-minute trivia <37 begin> */
  if (hashPrime <= memChunksMax) {
    panic(errprint0("Config error: hashprime must exceed memchunksmax"));
  }
  memHash = calloc(hashPrime + 1, sizeof(ChunkNode));
  if (memHash == NULL) {
    panic(errprint0("Cannot allocate the memory hash table"));
  }
  memHash[0].chunk = calloc(1 << 13, sizeof(Octa));
  if (memHash[0].chunk == NULL) {
    panic(errprint0("Cannot allocate memory chunk 0"));
  }
  memHash[hashPrime].chunk = calloc(1 << 13, sizeof(Octa));
  if (memHash[hashPrime].chunk == NULL) {
    panic(errprint0("Cannot allocate 0 memory chunk"));
  }
  memChunks = 1;
  fetchBot = calloc(fetchBufSize + 1, sizeof(Fetch));
  if (fetchBot == NULL) {
    panic(errprint0("Cannot allocate the fetch buffer"));
  }
  fetchTop = fetchBot + fetchBufSize;
  reorderBot = calloc(reorderBufSize + 1, sizeof(Control));
  if (reorderBot == NULL) {
    panic(errprint0("Cannot allocate the reorder buffer"));
  }
  reorderTop = reorderBot + reorderBufSize;
  wbufBot = calloc(writeBufSize + 1, sizeof(WriteNode));
  if (wbufBot == NULL) {
    panic(errprint0("Cannot allocate the write buffer"));
  }
  wbufTop = wbufBot + writeBufSize;
  if (bpN == 0) {
    bpTable = NULL;
  } else {
    if (bpA + bpB + bpC >= 32) {
      panic(errprint0("Config error: branch table is >= 4 gigabytes"));
    }
    bpTable = calloc(1 << (bpA + bpB + bpC), sizeof(char));
    if (bpTable == NULL) {
      panic(errprint0("Cannot allocate the branch table"));
    }
  }
  l = calloc(lringSize, sizeof(SpecNode));
  if (l == NULL) {
    panic(errprint0("Cannot allocate local registers"));
  }
  j = busWords;
  if (Icache != NULL && Icache->bb > j) {
    j = Icache->bb;
  }
  fetched = calloc(j, sizeof(Octa));
  if (fetched == NULL) {
    panic(errprint0("Cannot allocate prefetch buffer"));
  }
  dispatchStat = calloc(dispatchMax + 1, sizeof(int));
  if (dispatchStat == NULL) {
    panic(errprint0("Cannot allocate dispatch counts"));
  }
  swPT = !hwPT;
  /* touch up last-minute trivia <37 end> */
}
