/*
 * asm.c -- assembler
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "arith.h"
#include "spreg.h"
#include "instr.h"
#include "asm.h"


#define MAX_TOKENS	10


static char *msgs[] = {
  /* 0 */  "too many tokens on line",
  /* 1 */  "empty line",
  /* 2 */  "unknown instruction name",
  /* 3 */  "excess tokens on line",
  /* 4 */  "too few operands",
  /* 5 */  "illegal register",
  /* 6 */  "illegal immediate value",
  /* 7 */  "immediate value out of range",
  /* 8 */  "target cannot be reached",
  /* 9 */  "illegal special register"
};


static Bool getReg(char *str, int *reg) {
  char *end;

  if (*str != '$') {
    return FALSE;
  }
  *reg = strtoul(str + 1, &end, 10);
  if (*end != '\0') {
    return FALSE;
  }
  if (*reg < 0 || *reg >= 256) {
    return FALSE;
  }
  return TRUE;
}


static Bool getNum(char *str, Octa *val) {
  Octa base;
  int digVal;

  if (*str == '#') {
    base.h = 0;
    base.l = 16;
    str++;
    if (!isxdigit(*str)) {
      return FALSE;
    }
  } else {
    base.h = 0;
    base.l = 10;
    if (!isdigit(*str)) {
      return FALSE;
    }
  }
  *val = zero;
  while (isxdigit(*str)) {
    if (*str >= '0' && *str <= '9') {
      digVal = *str - '0';
    } else
    if (*str >= 'A' && *str <= 'F') {
      digVal = *str - 'A' + 10;
    } else
    if (*str >= 'a' && *str <= 'f') {
      digVal = *str - 'a' + 10;
    }
    if (digVal >= base.l) {
      return FALSE;
    }
    *val = incOcta(umulOcta(*val, base), digVal);
    str++;
  }
  return *str == '\0';
}


char *asmInstr(char *line, Octa locus, Tetra *instrPtr) {
  char *tokens[MAX_TOKENS];
  int n;
  char *p;
  Instr *instr;
  Tetra result;
  Octa tmp;
  int i1, i2, i3;
  int r1, r2, r3;
  SpReg *rp;

  /* separate tokens */
  n = 0;
  p = strtok(line, " \t\n,");
  while (p != NULL) {
    if (n == MAX_TOKENS) {
      return msgs[0];
    }
    tokens[n++] = p;
    p = strtok(NULL, " \t\n,");
  }
  if (n == 0) {
    return msgs[1];
  }
  /* lookup mnemonic */
  instr = lookupInstr(tokens[0]);
  if (instr == NULL) {
    return msgs[2];
  }
  /* do processing according to format */
  switch (instr->format) {
    case FMT_I8I8I8:
      if (n > 4) {
        return msgs[3];
      }
      if (n < 4) {
        return msgs[4];
      }
      if (!getNum(tokens[1], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i1 = tmp.l;
      if (!getNum(tokens[2], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i2 = tmp.l;
      if (!getNum(tokens[3], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i3 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (i1 << 16) |
               (i2 <<  8) |
               (i3 <<  0);
      break;
    case FMT_RRR:
      if (n > 4) {
        return msgs[3];
      }
      if (n < 4) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        return msgs[5];
      }
      if (!getReg(tokens[2], &r2)) {
        return msgs[5];
      }
      if (!getReg(tokens[3], &r3)) {
        return msgs[5];
      }
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (r2 <<  8) |
               (r3 <<  0);
      break;
    case FMT_RRI8:
      if (n > 4) {
        return msgs[3];
      }
      if (n < 4) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        return msgs[5];
      }
      if (!getReg(tokens[2], &r2)) {
        return msgs[5];
      }
      if (!getNum(tokens[3], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i1 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (r2 <<  8) |
               (i1 <<  0);
      break;
    case FMT_RI8R:
      if (n > 4) {
        return msgs[3];
      }
      if (n < 4) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        return msgs[5];
      }
      if (!getNum(tokens[2], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i1 = tmp.l;
      if (!getReg(tokens[3], &r2)) {
        return msgs[5];
      }
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (i1 <<  8) |
               (r2 <<  0);
      break;
    case FMT_RI8I8:
      if (n > 4) {
        return msgs[3];
      }
      if (n < 4) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        return msgs[5];
      }
      if (!getNum(tokens[2], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i1 = tmp.l;
      if (!getNum(tokens[3], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i2 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (i1 <<  8) |
               (i2 <<  0);
      break;
    case FMT_I8RR:
      if (n > 4) {
        return msgs[3];
      }
      if (n < 4) {
        return msgs[4];
      }
      if (!getNum(tokens[1], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i1 = tmp.l;
      if (!getReg(tokens[2], &r1)) {
        return msgs[5];
      }
      if (!getReg(tokens[3], &r2)) {
        return msgs[5];
      }
      result = ((Tetra) instr->opcode << 24) |
               (i1 << 16) |
               (r1 <<  8) |
               (r2 <<  0);
      break;
    case FMT_I8RI8:
      if (n > 4) {
        return msgs[3];
      }
      if (n < 4) {
        return msgs[4];
      }
      if (!getNum(tokens[1], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i1 = tmp.l;
      if (!getReg(tokens[2], &r1)) {
        return msgs[5];
      }
      if (!getNum(tokens[3], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i2 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (i1 << 16) |
               (r1 <<  8) |
               (i2 <<  0);
      break;
    case FMT_RI16:
      if (n > 3) {
        return msgs[3];
      }
      if (n < 3) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        return msgs[5];
      }
      if (!getNum(tokens[2], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x10000) {
        return msgs[7];
      }
      i1 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (i1 <<  0);
      break;
    case FMT_I8I16:
      if (n > 3) {
        return msgs[3];
      }
      if (n < 3) {
        return msgs[4];
      }
      if (!getNum(tokens[1], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i1 = tmp.l;
      if (!getNum(tokens[2], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x10000) {
        return msgs[7];
      }
      i2 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (i1 << 16) |
               (i2 <<  0);
      break;
    case FMT_RF16:
      if (n > 3) {
        return msgs[3];
      }
      if (n < 3) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        return msgs[5];
      }
      if (!getNum(tokens[2], &tmp)) {
        return msgs[6];
      }
      tmp = sarOcta(subOcta(tmp, locus), 2);
      if (tmp.h != 0x00000000 || tmp.l > 0x0000FFFF) {
        return msgs[8];
      }
      i1 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (i1 <<  0);
      break;
    case FMT_RB16:
      if (n > 3) {
        return msgs[3];
      }
      if (n < 3) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        return msgs[5];
      }
      if (!getNum(tokens[2], &tmp)) {
        return msgs[6];
      }
      tmp = sarOcta(subOcta(tmp, locus), 2);
      if (tmp.h != 0xFFFFFFFF || tmp.l < 0xFFFF0000) {
        return msgs[8];
      }
      i1 = tmp.l + 0x10000;
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (i1 <<  0);
      break;
    case FMT_F24:
      if (n > 2) {
        return msgs[3];
      }
      if (n < 2) {
        return msgs[4];
      }
      if (!getNum(tokens[1], &tmp)) {
        return msgs[6];
      }
      tmp = sarOcta(subOcta(tmp, locus), 2);
      if (tmp.h != 0x00000000 || tmp.l > 0x00FFFFFF) {
        return msgs[8];
      }
      i1 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (i1 <<  0);
      break;
    case FMT_B24:
      if (n > 2) {
        return msgs[3];
      }
      if (n < 2) {
        return msgs[4];
      }
      if (!getNum(tokens[1], &tmp)) {
        return msgs[6];
      }
      tmp = sarOcta(subOcta(tmp, locus), 2);
      if (tmp.h != 0xFFFFFFFF || tmp.l < 0xFF000000) {
        return msgs[8];
      }
      i1 = tmp.l + 0x1000000;
      result = ((Tetra) instr->opcode << 24) |
               (i1 <<  0);
      break;
    case FMT_RZ8S8:
      if (n > 3) {
        return msgs[3];
      }
      if (n < 3) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        return msgs[5];
      }
      if (!getReg(tokens[2], &r2)) {
        rp = lookupSpReg(tokens[2]);
        if (rp == NULL) {
          return msgs[9];
        }
        r2 = rp->code;
      }
      if (r2 >= 32) {
        return msgs[9];
      }
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (r2 <<  0);
      break;
    case FMT_S8Z8R:
      if (n > 3) {
        return msgs[3];
      }
      if (n < 3) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        rp = lookupSpReg(tokens[1]);
        if (rp == NULL) {
          return msgs[9];
        }
        r1 = rp->code;
      }
      if (r1 >= 32) {
        return msgs[9];
      }
      if (!getReg(tokens[2], &r2)) {
        return msgs[5];
      }
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (r2 <<  0);
      break;
    case FMT_S8Z8I8:
      if (n > 3) {
        return msgs[3];
      }
      if (n < 3) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        rp = lookupSpReg(tokens[1]);
        if (rp == NULL) {
          return msgs[9];
        }
        r1 = rp->code;
      }
      if (r1 >= 32) {
        return msgs[9];
      }
      if (!getNum(tokens[2], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i1 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (i1 <<  0);
      break;
    case FMT_RZ8R:
      if (n > 3) {
        return msgs[3];
      }
      if (n < 3) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        return msgs[5];
      }
      if (!getReg(tokens[2], &r2)) {
        return msgs[5];
      }
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (r2 <<  0);
      break;
    case FMT_RZ8I8:
      if (n > 3) {
        return msgs[3];
      }
      if (n < 3) {
        return msgs[4];
      }
      if (!getReg(tokens[1], &r1)) {
        return msgs[5];
      }
      if (!getNum(tokens[2], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i1 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (r1 << 16) |
               (i1 <<  0);
      break;
    case FMT_Z8Z8I8:
      if (n > 2) {
        return msgs[3];
      }
      if (n < 2) {
        return msgs[4];
      }
      if (!getNum(tokens[1], &tmp)) {
        return msgs[6];
      }
      if (tmp.h != 0 || tmp.l >= 0x100) {
        return msgs[7];
      }
      i1 = tmp.l;
      result = ((Tetra) instr->opcode << 24) |
               (i1 <<  0);
      break;
    default:
      error("unknown instruction format");
  }
  /* line successfully assembled */
  *instrPtr = result;
  return NULL;
}
