/*
 * cpu.c -- CPU simulation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "except.h"
#include "arith.h"
#include "spreg.h"
#include "instr.h"
#include "cpu.h"
#include "mmu.h"
#include "timer.h"

#include "abstime.h"


#define RESUME_NONE	0	/* actually no resume at all */
#define RESUME_0	1	/* RESUME 0 */
#define RESUME_1	2	/* RESUME 1 */

#define ROP_AGAIN	0	/* repeat the command in rX */
#define ROP_CONT	1	/* same as 0, but substitute operands */
#define ROP_SET		2	/* set r[X] to rZ */
#define ROP_MAP		3	/* same as 0, but install a VAT mapping */


/**************************************************************/


static Bool debugStack = FALSE;
static Bool debugInstr = TRUE;
static Bool debugTrips = TRUE;
static Bool debugTraps = TRUE;

static int resuming;		/* one of the above RESUME_x codes */
static int ropcode;		/* one of the above ROP_x codes */
static Tetra rexcept;		/* exceptions of a resumed instruction */

static Octa ip;			/* instruction pointer */
static Octa next;		/* value to update ip to */
static Octa altNext;		/* alternative 'next' value for jumps */

static Tetra instr;		/* current instruction */
static int opcode;		/* opcode of current instruction */
static int x, y, z;		/* operand fields of current instruction */
static Byte opflags;		/* bit-coded type of current instruction */
static Octa opX, opY, opZ;	/* operands of current instruction */
static Octa sumYZ;		/* opY + opZ */
static Octa res;		/* result of current instruction */
static Octa *resPtr;		/* where to store the result */

static Octa gr[GR_NUM];		/* global registers */
static Octa lr[LR_NUM];		/* local registers */

static int G;			/* easily accessible copy of rG */
static int L;			/* easily accessible copy of rL */
static int O;			/* congruent to (rO/8) mod LR_NUM */
static int S;			/* congruent to (rS/8) mod LR_NUM */

static Bool breakSet;		/* true iff CPU should break */
static Octa breakAddr;		/* loaction where it should break */

static int instrCount;		/* count to tick the timer */

static Tetra tripsPending;	/* pending trips are noticed here */
static Tetra ftrapsPending;	/* pending forced traps are noticed here */
static Octa dtrapsPending;	/* pending dynamic traps are noticed here */

static Bool run;		/* used to halt the running CPU */


/**************************************************************/


static void stackStore(void) {
  int gamma;

  gamma = S & LR_MASK;
  mmuWriteOcta(gr[rS], lr[gamma]);
  if (debugStack) {
    cPrintf("STACK store:  M8[#%08X%08X] = l[%d] (#%08X%08X), rS += 8\n",
            gr[rS].h, gr[rS].l, gamma, lr[gamma].h, lr[gamma].l);
  }
  gr[rS] = incOcta(gr[rS], 8);
  S++;
}


static void stackLoad(void) {
  int gamma;

  S--;
  gr[rS] = incOcta(gr[rS], -8);
  gamma = S & LR_MASK;
  lr[gamma] = mmuReadOcta(gr[rS]);
  if (debugStack) {
    cPrintf("STACK load:   rS -= 8, l[%d] = M8[#%08X%08X] (#%08X%08X)\n",
            gamma, gr[rS].h, gr[rS].l, lr[gamma].h, lr[gamma].l);
  }
}


/**************************************************************/


static Octa readDR(int reg) {
  if (reg >= G) {
    /* global register */
    return gr[reg];
  }
  if (reg < L) {
    /* local register */
    return lr[(O + reg) & LR_MASK];
  }
  /* marginal register */
  return zero;
}


static Octa *writeDR(int reg) {
  if (reg >= G) {
    /* global register */
    return &gr[reg];
  }
  /* marginal or local register */
  while (reg >= L) {
    /* increase rL */
    lr[(O + L) & LR_MASK] = zero;
    L++;
    gr[rL].l = L;
    if (debugStack) {
      cPrintf("STACK incr:   beta = %d, rL = %d\n",
              (O + L) & LR_MASK, L);
    }
    if (((S - O - L) & LR_MASK) == 0) {
      stackStore();
    }
  }
  /* reg is now a local register */
  return &lr[(O + reg) & LR_MASK];
}


/**************************************************************/


static Octa readSpecial(int reg) {
  /* every special register is readable */
  return gr[reg];
}


static void writeSpecial(int reg, Octa data) {
  int i;
  int s, n;

  /* writing to certain special registers is restricted */
  if (reg >= 8 && reg <= 11) {
    /* nobody is allowed to write to these registers: */
    /* rC, rN, rO, rS */
    throwException(EXC_DT_B);
  } else
  if (reg >= 12 && reg <= 18) {
    /* only the operating system may write to these registers: */
    /* rI, rT, rTT, rK, rQ, rU, rV */
    if ((ip.h & SIGN(32)) == 0) {
      throwException(EXC_DT_K);
    } else {
      gr[reg] = data;
      if (reg == rV) {
        s = (data.h >> 8) & 0x00FF;
        n = (data.l >> 3) & 0x03FF;
        mmuCtrl(s, n);
      }
    }
  } else
  if (reg == rA) {
    /* only the lower 18 bits may change */
    if (data.h != 0 || data.l >= 0x40000) {
      throwException(EXC_DT_B);
    } else {
      gr[rA] = data;
    }
  } else
  if (reg == rL) {
    /* set rL to min(data,rL) */
    if (data.h == 0 && data.l < L) {
      L = data.l;
      gr[rL].l = L;
    }
  } else
  if (reg == rG) {
    /* value must be in range 32..255 and not less than rL */
    if (data.h != 0 || data.l > 255 || data.l < L || data.l < 32) {
      throwException(EXC_DT_B);
    } else {
      for (i = data.l; i < G; i++) {
        gr[i] = zero;
      }
      G = data.l;
      gr[rG].l = G;
    }
  } else {
    /* the remaining registers may be written to without restrictions */
    gr[reg] = data;
  }
}


/**************************************************************/


static void push(int reg) {
  Octa holeData;
  Octa *holeAddr;

  if (reg >= G) {
    reg = L;
  }
  holeData.h = 0;
  holeData.l = reg;
  holeAddr = writeDR(reg);
  *holeAddr = holeData;
  O += reg + 1;
  gr[rO] = incOcta(gr[rO], (reg + 1) << 3);
  L -= reg + 1;
  gr[rL].l = L;
  if (debugStack) {
    cPrintf("STACK push:   l[%d] = %d, alpha = %d, rL = %d\n",
            (O - 1) & LR_MASK, holeData.l, O & LR_MASK, L);
  }
}


static void pop(int reg) {
  Octa holeData;
  int numRegs;

  if (reg != 0 && reg <= L) {
    holeData = lr[(O + reg - 1) & LR_MASK];
  }
  if (gr[rS].l == gr[rO].l) {
    stackLoad();
  }
  numRegs = lr[(O - 1) & LR_MASK].l & 0xFF;
  while ((Tetra) (O - S) <= (Tetra) numRegs) {
    stackLoad();
  }
  L = numRegs + (reg <= L ? reg : L + 1);
  if (L > G) {
    L = G;
  }
  if (L > numRegs) {
    lr[(O - 1) & LR_MASK] = holeData;
  }
  gr[rL].l = L;
  O -= numRegs + 1;
  gr[rO] = incOcta(gr[rO], -((numRegs + 1) << 3));
  if (debugStack) {
    if (L > numRegs) {
      cPrintf("STACK pop:    l[%d] = %d, alpha = %d, rL = %d\n",
              (O + numRegs) & LR_MASK, holeData.l, O & LR_MASK, L);
    } else {
      cPrintf("STACK pop:    alpha = %d, rL = %d\n",
              O & LR_MASK, L);
    }
  }
}


/**************************************************************/


static void handleRealTimeTasks(void) {
  /* handle 'real-time' tasks */
  if (++instrCount == INSTRS_PER_MSEC) {
    instrCount = 0;
    timerTick();
  }
}


/**************************************************************/


static void handleTraps(void) {
  Tetra mask;
  int i;

  if (ftrapsPending == 0 &&
      dtrapsPending.h == 0 && dtrapsPending.l == 0) {
    /* no trap is pending */
    return;
  }
  if (ftrapsPending != 0) {
    /* exactly one forced trap is pending */
    if ((ftrapsPending & (1 << (-EXC_FT_TRAP - 1))) != 0) {
      /* trap instruction */
      if (debugTraps) {
        cPrintf("TRAP: triggered by trap instruction\n");
      }
      ftrapsPending &= ~(1 << (-EXC_FT_TRAP - 1));
      gr[rXX].h = 0x80000000;
      gr[rXX].l = instr;
      gr[rYY] = opY;
      gr[rZZ] = opZ;
    } else
    if ((ftrapsPending & (1 << (-EXC_FT_EMUL - 1))) != 0) {
      /* emulated instruction */
      if (debugTraps) {
        cPrintf("TRAP: triggered by emulated instruction\n");
      }
      ftrapsPending &= ~(1 << (-EXC_FT_EMUL - 1));
      gr[rXX].h = 0x02000000;
      gr[rXX].l = instr;
      gr[rYY] = opY;
      gr[rZZ] = opZ;
    } else
    if ((ftrapsPending & (1 << (-EXC_FT_VAT - 1))) != 0) {
      /* virtual address translation */
      if (debugTraps) {
        cPrintf("TRAP: triggered by virtual address translation\n");
      }
      ftrapsPending &= ~(1 << (-EXC_FT_VAT - 1));
      gr[rXX].h = 0x03000000;
      gr[rXX].l = instr;
      gr[rYY] = sumYZ;
    } else {
      error("unknown forced trap triggered");
    }
    gr[rK] = zero;
    gr[rWW] = next;
    next = gr[rT];
    gr[rBB] = gr[255];
    gr[255] = gr[rJ];
    ip = next;
  } else {
    /* at least one dynamic trap is pending */
    if (dtrapsPending.h != 0) {
      mask = 1;
      for (i = 0; i < 32; i++) {
        if ((dtrapsPending.h & mask) != 0) {
          break;
        }
        mask <<= 1;
      }
      i += 32;
    } else {
      mask = 1;
      for (i = 0; i < 32; i++) {
        if ((dtrapsPending.l & mask) != 0) {
          break;
        }
        mask <<= 1;
      }
    }
    if (debugTraps) {
      cPrintf("TRAP: dynamic trap %d triggered\n", i);
    }
  }
}


/**************************************************************/


static void emu(void) {
  throwException(EXC_FT_EMUL);
}


/*------------------------------------------------------------*/

/* 00 - 0F */


static void TRAP(void) {
  /* trap */
  throwException(EXC_FT_TRAP);
}


/*------------------------------------------------------------*/

/* 10 - 1F */


static void MUL(void) {
  /* multiply */
  res = mulOcta(opY, opZ);
  *resPtr = res;
  if (overflow) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void MULI(void) {
  /* multiply immediate */
  res = mulOcta(opY, opZ);
  *resPtr = res;
  if (overflow) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void MULU(void) {
  /* multiply unsigned */
  res = umulOcta(opY, opZ);
  *resPtr = res;
  gr[rH] = auxOut;
}


static void MULUI(void) {
  /* multiply unsigned immediate */
  res = umulOcta(opY, opZ);
  *resPtr = res;
  gr[rH] = auxOut;
}


static void DIV(void) {
  /* divide */
  if (opZ.h == 0 && opZ.l == 0) {
    /* integer divide check */
    res = zero;
    *resPtr = res;
    gr[rR] = opY;
    cpuSetTrip(TRIP_D_BIT);
  } else {
    res = divOcta(opY, opZ);
    *resPtr = res;
    gr[rR] = auxOut;
    if (overflow) {
      /* integer overflow */
      cpuSetTrip(TRIP_V_BIT);
    }
  }
}


static void DIVI(void) {
  /* divide immediate */
  if (opZ.h == 0 && opZ.l == 0) {
    /* integer divide check */
    res = zero;
    *resPtr = res;
    gr[rR] = opY;
    cpuSetTrip(TRIP_D_BIT);
  } else {
    res = divOcta(opY, opZ);
    *resPtr = res;
    gr[rR] = auxOut;
    if (overflow) {
      /* integer overflow */
      cpuSetTrip(TRIP_V_BIT);
    }
  }
}


static void DIVU(void) {
  /* divide unsigned */
  res = udivOcta(gr[rD], opY, opZ);
  *resPtr = res;
  gr[rR] = auxOut;
}


static void DIVUI(void) {
  /* divide unsigned immediate */
  res = udivOcta(gr[rD], opY, opZ);
  *resPtr = res;
  gr[rR] = auxOut;
}


/*------------------------------------------------------------*/

/* 20 - 2F */


static void ADD(void) {
  /* add */
  res = addOcta(opY, opZ);
  *resPtr = res;
  if (((opY.h ^ opZ.h) & SIGN(32)) == 0 &&
      ((opY.h ^ res.h) & SIGN(32)) != 0) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void ADDI(void) {
  /* add immediate */
  res = addOcta(opY, opZ);
  *resPtr = res;
  if (((opY.h ^ opZ.h) & SIGN(32)) == 0 &&
      ((opY.h ^ res.h) & SIGN(32)) != 0) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void ADDU(void) {
  /* add unsigned */
  res = addOcta(opY, opZ);
  *resPtr = res;
}


static void ADDUI(void) {
  /* add unsigned immediate */
  res = addOcta(opY, opZ);
  *resPtr = res;
}


static void SUB(void) {
  /* subtract */
  res = subOcta(opY, opZ);
  *resPtr = res;
  if (((res.h ^ opZ.h) & SIGN(32)) == 0 &&
      ((res.h ^ opY.h) & SIGN(32)) != 0) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void SUBI(void) {
  /* subtract immediate */
  res = subOcta(opY, opZ);
  *resPtr = res;
  if (((res.h ^ opZ.h) & SIGN(32)) == 0 &&
      ((res.h ^ opY.h) & SIGN(32)) != 0) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void SUBU(void) {
  /* subtract unsigned */
  res = subOcta(opY, opZ);
  *resPtr = res;
}


static void SUBUI(void) {
  /* subtract unsigned immediate */
  res = subOcta(opY, opZ);
  *resPtr = res;
}


static void ADDU2(void) {
  /* times 2 and add unsigned */
  res = addOcta(shlOcta(opY, 1), opZ);
  *resPtr = res;
}


static void ADDUI2(void) {
  /* times 2 and add unsigned immediate */
  res = addOcta(shlOcta(opY, 1), opZ);
  *resPtr = res;
}


static void ADDU4(void) {
  /* times 4 and add unsigned */
  res = addOcta(shlOcta(opY, 2), opZ);
  *resPtr = res;
}


static void ADDUI4(void) {
  /* times 4 and add unsigned immediate */
  res = addOcta(shlOcta(opY, 2), opZ);
  *resPtr = res;
}


static void ADDU8(void) {
  /* times 8 and add unsigned */
  res = addOcta(shlOcta(opY, 3), opZ);
  *resPtr = res;
}


static void ADDUI8(void) {
  /* times 8 and add unsigned immediate */
  res = addOcta(shlOcta(opY, 3), opZ);
  *resPtr = res;
}


static void ADDU16(void) {
  /* times 16 and add unsigned */
  res = addOcta(shlOcta(opY, 4), opZ);
  *resPtr = res;
}


static void ADDUI16(void) {
  /* times 16 and add unsigned immediate */
  res = addOcta(shlOcta(opY, 4), opZ);
  *resPtr = res;
}


/*------------------------------------------------------------*/

/* 30 - 3F */


static void CMP(void) {
  /* compare */
  if ((opY.h & SIGN(32)) > (opZ.h & SIGN(32))) {
    res = negOne;
  } else
  if ((opY.h & SIGN(32)) < (opZ.h & SIGN(32))) {
    res = one;
  } else
  if (opY.h < opZ.h) {
    res = negOne;
  } else
  if (opY.h > opZ.h) {
    res = one;
  } else
  if (opY.l < opZ.l) {
    res = negOne;
  } else
  if (opY.l > opZ.l) {
    res = one;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void CMPI(void) {
  /* compare immediate */
  if ((opY.h & SIGN(32)) > (opZ.h & SIGN(32))) {
    res = negOne;
  } else
  if ((opY.h & SIGN(32)) < (opZ.h & SIGN(32))) {
    res = one;
  } else
  if (opY.h < opZ.h) {
    res = negOne;
  } else
  if (opY.h > opZ.h) {
    res = one;
  } else
  if (opY.l < opZ.l) {
    res = negOne;
  } else
  if (opY.l > opZ.l) {
    res = one;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void CMPU(void) {
  /* compare unsigned */
  if (opY.h < opZ.h) {
    res = negOne;
  } else
  if (opY.h > opZ.h) {
    res = one;
  } else
  if (opY.l < opZ.l) {
    res = negOne;
  } else
  if (opY.l > opZ.l) {
    res = one;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void CMPUI(void) {
  /* compare unsigned immediate */
  if (opY.h < opZ.h) {
    res = negOne;
  } else
  if (opY.h > opZ.h) {
    res = one;
  } else
  if (opY.l < opZ.l) {
    res = negOne;
  } else
  if (opY.l > opZ.l) {
    res = one;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void NEG(void) {
  /* negate */
  res = subOcta(opY, opZ);
  *resPtr = res;
  if (((res.h ^ opZ.h) & SIGN(32)) == 0 &&
      ((res.h ^ opY.h) & SIGN(32)) != 0) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void NEGI(void) {
  /* negate immediate */
  res = subOcta(opY, opZ);
  *resPtr = res;
  if (((res.h ^ opZ.h) & SIGN(32)) == 0 &&
      ((res.h ^ opY.h) & SIGN(32)) != 0) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void NEGU(void) {
  /* negate unsigned */
  res = subOcta(opY, opZ);
  *resPtr = res;
}


static void NEGUI(void) {
  /* negate unsigned immediate */
  res = subOcta(opY, opZ);
  *resPtr = res;
}


static void SL(void) {
  /* shift left */
  int n;
  Octa tmp;

  if (opZ.h != 0 || opZ.l >= 64) {
    n = 64;
  } else {
    n = opZ.l;
  }
  res = shlOcta(opY, n);
  *resPtr = res;
  tmp = sarOcta(res, n);
  if (tmp.h != opY.h || tmp.l != opY.l) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void SLI(void) {
  /* shift left immediate */
  int n;
  Octa tmp;

  if (opZ.h != 0 || opZ.l >= 64) {
    n = 64;
  } else {
    n = opZ.l;
  }
  res = shlOcta(opY, n);
  *resPtr = res;
  tmp = sarOcta(res, n);
  if (tmp.h != opY.h || tmp.l != opY.l) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void SLU(void) {
  /* shift left unsigned */
  int n;

  if (opZ.h != 0 || opZ.l >= 64) {
    n = 64;
  } else {
    n = opZ.l;
  }
  res = shlOcta(opY, n);
  *resPtr = res;
}


static void SLUI(void) {
  /* shift left unsigned immediate */
  int n;

  if (opZ.h != 0 || opZ.l >= 64) {
    n = 64;
  } else {
    n = opZ.l;
  }
  res = shlOcta(opY, n);
  *resPtr = res;
}


static void SR(void) {
  /* shift right */
  int n;

  if (opZ.h != 0 || opZ.l >= 64) {
    n = 64;
  } else {
    n = opZ.l;
  }
  res = sarOcta(opY, n);
  *resPtr = res;
}


static void SRI(void) {
  /* shift right immediate */
  int n;

  if (opZ.h != 0 || opZ.l >= 64) {
    n = 64;
  } else {
    n = opZ.l;
  }
  res = sarOcta(opY, n);
  *resPtr = res;
}


static void SRU(void) {
  /* shift right unsigned */
  int n;

  if (opZ.h != 0 || opZ.l >= 64) {
    n = 64;
  } else {
    n = opZ.l;
  }
  res = shrOcta(opY, n);
  *resPtr = res;
}


static void SRUI(void) {
  /* shift right unsigned immediate */
  int n;

  if (opZ.h != 0 || opZ.l >= 64) {
    n = 64;
  } else {
    n = opZ.l;
  }
  res = shrOcta(opY, n);
  *resPtr = res;
}


/*------------------------------------------------------------*/

/* 40 - 4F */


static void BN(void) {
  /* branch if negative */
  if ((opX.h & SIGN(32)) != 0) {
    next = altNext;
  }
}


static void BNB(void) {
  /* branch if negative back */
  if ((opX.h & SIGN(32)) != 0) {
    next = altNext;
  }
}


static void BZ(void) {
  /* branch if zero */
  if (opX.h == 0 && opX.l == 0) {
    next = altNext;
  }
}


static void BZB(void) {
  /* branch if zero back */
  if (opX.h == 0 && opX.l == 0) {
    next = altNext;
  }
}


static void BP(void) {
  /* branch if positive */
  if ((opX.h & SIGN(32)) == 0 && (opX.h != 0 || opX.l != 0)) {
    next = altNext;
  }
}


static void BPB(void) {
  /* branch if positive back */
  if ((opX.h & SIGN(32)) == 0 && (opX.h != 0 || opX.l != 0)) {
    next = altNext;
  }
}


static void BOD(void) {
  /* branch if odd */
  if ((opX.l & 1) != 0) {
    next = altNext;
  }
}


static void BODB(void) {
  /* branch if odd back */
  if ((opX.l & 1) != 0) {
    next = altNext;
  }
}


static void BNN(void) {
  /* branch if nonnegative */
  if ((opX.h & SIGN(32)) == 0) {
    next = altNext;
  }
}


static void BNNB(void) {
  /* branch if nonnegative back */
  if ((opX.h & SIGN(32)) == 0) {
    next = altNext;
  }
}


static void BNZ(void) {
  /* branch if nonzero */
  if (opX.h != 0 || opX.l != 0) {
    next = altNext;
  }
}


static void BNZB(void) {
  /* branch if nonzero back */
  if (opX.h != 0 || opX.l != 0) {
    next = altNext;
  }
}


static void BNP(void) {
  /* branch if nonpositive */
  if ((opX.h & SIGN(32)) != 0 || (opX.h == 0 && opX.l == 0)) {
    next = altNext;
  }
}


static void BNPB(void) {
  /* branch if nonpositive back */
  if ((opX.h & SIGN(32)) != 0 || (opX.h == 0 && opX.l == 0)) {
    next = altNext;
  }
}


static void BEV(void) {
  /* branch if even */
  if ((opX.l & 1) == 0) {
    next = altNext;
  }
}


static void BEVB(void) {
  /* branch if even back */
  if ((opX.l & 1) == 0) {
    next = altNext;
  }
}


/*------------------------------------------------------------*/

/* 50 - 5F */


static void PBN(void) {
  /* probable branch if negative */
  if ((opX.h & SIGN(32)) != 0) {
    next = altNext;
  }
}


static void PBNB(void) {
  /* probable branch if negative back */
  if ((opX.h & SIGN(32)) != 0) {
    next = altNext;
  }
}


static void PBZ(void) {
  /* probable branch if zero */
  if (opX.h == 0 && opX.l == 0) {
    next = altNext;
  }
}


static void PBZB(void) {
  /* probable branch if zero back */
  if (opX.h == 0 && opX.l == 0) {
    next = altNext;
  }
}


static void PBP(void) {
  /* probable branch if positive */
  if ((opX.h & SIGN(32)) == 0 && (opX.h != 0 || opX.l != 0)) {
    next = altNext;
  }
}


static void PBPB(void) {
  /* probable branch if positive back */
  if ((opX.h & SIGN(32)) == 0 && (opX.h != 0 || opX.l != 0)) {
    next = altNext;
  }
}


static void PBOD(void) {
  /* probable branch if odd */
  if ((opX.l & 1) != 0) {
    next = altNext;
  }
}


static void PBODB(void) {
  /* probable branch if odd back */
  if ((opX.l & 1) != 0) {
    next = altNext;
  }
}


static void PBNN(void) {
  /* probable branch if nonnegative */
  if ((opX.h & SIGN(32)) == 0) {
    next = altNext;
  }
}


static void PBNNB(void) {
  /* probable branch if nonnegative back */
  if ((opX.h & SIGN(32)) == 0) {
    next = altNext;
  }
}


static void PBNZ(void) {
  /* probable branch if nonzero */
  if (opX.h != 0 || opX.l != 0) {
    next = altNext;
  }
}


static void PBNZB(void) {
  /* probable branch if nonzero back */
  if (opX.h != 0 || opX.l != 0) {
    next = altNext;
  }
}


static void PBNP(void) {
  /* probable branch if nonpositive */
  if ((opX.h & SIGN(32)) != 0 || (opX.h == 0 && opX.l == 0)) {
    next = altNext;
  }
}


static void PBNPB(void) {
  /* probable branch if nonpositive back */
  if ((opX.h & SIGN(32)) != 0 || (opX.h == 0 && opX.l == 0)) {
    next = altNext;
  }
}


static void PBEV(void) {
  /* probable branch if even */
  if ((opX.l & 1) == 0) {
    next = altNext;
  }
}


static void PBEVB(void) {
  /* probable branch if even back */
  if ((opX.l & 1) == 0) {
    next = altNext;
  }
}


/*------------------------------------------------------------*/

/* 60 - 6F */


static void CSN(void) {
  /* conditionally set if negative */
  if ((opY.h & SIGN(32)) != 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSNI(void) {
  /* conditionally set if negative immediate */
  if ((opY.h & SIGN(32)) != 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSZ(void) {
  /* conditionally set if zero */
  if (opY.h == 0 && opY.l == 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSZI(void) {
  /* conditionally set if zero immediate */
  if (opY.h == 0 && opY.l == 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSP(void) {
  /* conditionally set if positive */
  if ((opY.h & SIGN(32)) == 0 && (opY.h != 0 || opY.l != 0)) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSPI(void) {
  /* conditionally set if positive immediate */
  if ((opY.h & SIGN(32)) == 0 && (opY.h != 0 || opY.l != 0)) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSOD(void) {
  /* conditionally set if odd */
  if ((opY.l & 1) != 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSODI(void) {
  /* conditionally set if odd immediate */
  if ((opY.l & 1) != 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSNN(void) {
  /* conditionally set if nonnegative */
  if ((opY.h & SIGN(32)) == 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSNNI(void) {
  /* conditionally set if nonnegative immediate */
  if ((opY.h & SIGN(32)) == 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSNZ(void) {
  /* conditionally set if nonzero */
  if (opY.h != 0 || opY.l != 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSNZI(void) {
  /* conditionally set if nonzero immediate */
  if (opY.h != 0 || opY.l != 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSNP(void) {
  /* conditionally set if nonpositive */
  if ((opY.h & SIGN(32)) != 0 || (opY.h == 0 && opY.l == 0)) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSNPI(void) {
  /* conditionally set if nonpositive immediate */
  if ((opY.h & SIGN(32)) != 0 || (opY.h == 0 && opY.l == 0)) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSEV(void) {
  /* conditionally set if even */
  if ((opY.l & 1) == 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


static void CSEVI(void) {
  /* conditionally set if even immediate */
  if ((opY.l & 1) == 0) {
    res = opZ;
  } else {
    res = opX;
  }
  *resPtr = res;
}


/*------------------------------------------------------------*/

/* 70 - 7F */


static void ZSN(void) {
  /* zero or set if negative */
  if ((opY.h & SIGN(32)) != 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSNI(void) {
  /* zero or set if negative immediate */
  if ((opY.h & SIGN(32)) != 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSZ(void) {
  /* zero or set if zero */
  if (opY.h == 0 && opY.l == 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSZI(void) {
  /* zero or set if zero immediate */
  if (opY.h == 0 && opY.l == 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSP(void) {
  /* zero or set if positive */
  if ((opY.h & SIGN(32)) == 0 && (opY.h != 0 || opY.l != 0)) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSPI(void) {
  /* zero or set if positive immediate */
  if ((opY.h & SIGN(32)) == 0 && (opY.h != 0 || opY.l != 0)) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSOD(void) {
  /* zero or set if odd */
  if ((opY.l & 1) != 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSODI(void) {
  /* zero or set if odd immediate */
  if ((opY.l & 1) != 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSNN(void) {
  /* zero or set if nonnegative */
  if ((opY.h & SIGN(32)) == 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSNNI(void) {
  /* zero or set if nonnegative immediate */
  if ((opY.h & SIGN(32)) == 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSNZ(void) {
  /* zero or set if nonzero */
  if (opY.h != 0 || opY.l != 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSNZI(void) {
  /* zero or set if nonzero immediate */
  if (opY.h != 0 || opY.l != 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSNP(void) {
  /* zero or set if nonpositive */
  if ((opY.h & SIGN(32)) != 0 || (opY.h == 0 && opY.l == 0)) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSNPI(void) {
  /* zero or set if nonpositive immediate */
  if ((opY.h & SIGN(32)) != 0 || (opY.h == 0 && opY.l == 0)) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSEV(void) {
  /* zero or set if even */
  if ((opY.l & 1) == 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


static void ZSEVI(void) {
  /* zero or set if even immediate */
  if ((opY.l & 1) == 0) {
    res = opZ;
  } else {
    res = zero;
  }
  *resPtr = res;
}


/*------------------------------------------------------------*/

/* 80 - 8F */


static void LDB(void) {
  /* load byte */
  res.l = mmuReadByte(sumYZ);
  res.l = SEXT(res.l, 8);
  res.h = (res.l & SIGN(32) ? -1 : 0);
  *resPtr = res;
}


static void LDBI(void) {
  /* load byte immediate */
  res.l = mmuReadByte(sumYZ);
  res.l = SEXT(res.l, 8);
  res.h = (res.l & SIGN(32) ? -1 : 0);
  *resPtr = res;
}


static void LDBU(void) {
  /* load byte unsigned */
  res.l = mmuReadByte(sumYZ);
  res.l = ZEXT(res.l, 8);
  res.h = 0;
  *resPtr = res;
}


static void LDBUI(void) {
  /* load byte unsigned immediate */
  res.l = mmuReadByte(sumYZ);
  res.l = ZEXT(res.l, 8);
  res.h = 0;
  *resPtr = res;
}


static void LDW(void) {
  /* load wyde */
  res.l = mmuReadWyde(sumYZ);
  res.l = SEXT(res.l, 16);
  res.h = (res.l & SIGN(32) ? -1 : 0);
  *resPtr = res;
}


static void LDWI(void) {
  /* load wyde immediate */
  res.l = mmuReadWyde(sumYZ);
  res.l = SEXT(res.l, 16);
  res.h = (res.l & SIGN(32) ? -1 : 0);
  *resPtr = res;
}


static void LDWU(void) {
  /* load wyde unsigned */
  res.l = mmuReadWyde(sumYZ);
  res.l = ZEXT(res.l, 16);
  res.h = 0;
  *resPtr = res;
}


static void LDWUI(void) {
  /* load wyde unsigned immediate */
  res.l = mmuReadWyde(sumYZ);
  res.l = ZEXT(res.l, 16);
  res.h = 0;
  *resPtr = res;
}


static void LDT(void) {
  /* load tetra */
  res.l = mmuReadTetra(sumYZ);
  res.h = (res.l & SIGN(32) ? -1 : 0);
  *resPtr = res;
}


static void LDTI(void) {
  /* load tetra immediate */
  res.l = mmuReadTetra(sumYZ);
  res.h = (res.l & SIGN(32) ? -1 : 0);
  *resPtr = res;
}


static void LDTU(void) {
  /* load tetra unsigned */
  res.l = mmuReadTetra(sumYZ);
  res.h = 0;
  *resPtr = res;
}


static void LDTUI(void) {
  /* load tetra unsigned immediate */
  res.l = mmuReadTetra(sumYZ);
  res.h = 0;
  *resPtr = res;
}


static void LDO(void) {
  /* load octa */
  res = mmuReadOcta(sumYZ);
  *resPtr = res;
}


static void LDOI(void) {
  /* load octa immediate */
  res = mmuReadOcta(sumYZ);
  *resPtr = res;
}


static void LDOU(void) {
  /* load octa unsigned */
  res = mmuReadOcta(sumYZ);
  *resPtr = res;
}


static void LDOUI(void) {
  /* load octa unsigned immediate */
  res = mmuReadOcta(sumYZ);
  *resPtr = res;
}


/*------------------------------------------------------------*/

/* 90 - 9F */


static void LDHT(void) {
  /* load high tetra */
  res.h = mmuReadTetra(sumYZ);
  res.l = 0;
  *resPtr = res;
}


static void LDHTI(void) {
  /* load high tetra immediate */
  res.h = mmuReadTetra(sumYZ);
  res.l = 0;
  *resPtr = res;
}


static void CSWAP(void) {
  /* compare and swap octabytes */
  Octa data;

  data = mmuReadOcta(sumYZ);
  if (data.h == gr[rP].h && data.l == gr[rP].l) {
    mmuWriteOcta(sumYZ, opX);
    res = one;
  } else {
    gr[rP] = data;
    res = zero;
  }
  *resPtr = res;
}


static void CSWAPI(void) {
  /* compare and swap octabytes immediate */
  Octa data;

  data = mmuReadOcta(sumYZ);
  if (data.h == gr[rP].h && data.l == gr[rP].l) {
    mmuWriteOcta(sumYZ, opX);
    res = one;
  } else {
    gr[rP] = data;
    res = zero;
  }
  *resPtr = res;
}


static void LDUNC(void) {
  /* load octa uncached */
  res = mmuReadOcta(sumYZ);
  *resPtr = res;
}


static void LDUNCI(void) {
  /* load octa uncached immediate */
  res = mmuReadOcta(sumYZ);
  *resPtr = res;
}


static void LDVTS(void) {
  /* load virtual translation status */
  if ((ip.h & SIGN(32)) != 0) {
    /* for operating system only */
    throwException(EXC_DT_K);
  }
  res.h = 0;
  res.l = mmuUpdate(sumYZ);
  *resPtr = res;
}


static void LDVTSI(void) {
  /* load virtual translation status immediate */
  if ((ip.h & SIGN(32)) != 0) {
    /* for operating system only */
    throwException(EXC_DT_K);
  }
  res.h = 0;
  res.l = mmuUpdate(sumYZ);
  *resPtr = res;
}


static void PRELD(void) {
  /* preload data */
  /* has no effect because caches are not simulated */
}


static void PRELDI(void) {
  /* preload data immediate */
  /* has no effect because caches are not simulated */
}


static void PREGO(void) {
  /* prefetch to go */
  /* has no effect because caches are not simulated */
}


static void PREGOI(void) {
  /* prefetch to go immediate */
  /* has no effect because caches are not simulated */
}


static void GO(void) {
  /* go to location */
  *resPtr = next;
  next = sumYZ;
}


static void GOI(void) {
  /* go to location immediate */
  *resPtr = next;
  next = sumYZ;
}


/*------------------------------------------------------------*/

/* A0 - AF */


static void STB(void) {
  /* store byte */
  Octa tmp;

  mmuWriteByte(sumYZ, (Byte) opX.l);
  tmp = sarOcta(shlOcta(opX, 56), 56);
  if (tmp.h != opX.h || tmp.l != opX.l) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void STBI(void) {
  /* store byte immediate */
  Octa tmp;

  mmuWriteByte(sumYZ, (Byte) opX.l);
  tmp = sarOcta(shlOcta(opX, 56), 56);
  if (tmp.h != opX.h || tmp.l != opX.l) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void STBU(void) {
  /* store byte unsigned */
  mmuWriteByte(sumYZ, (Byte) opX.l);
}


static void STBUI(void) {
  /* store byte unsigned immediate */
  mmuWriteByte(sumYZ, (Byte) opX.l);
}


static void STW(void) {
  /* store wyde */
  Octa tmp;

  mmuWriteWyde(sumYZ, (Wyde) opX.l);
  tmp = sarOcta(shlOcta(opX, 48), 48);
  if (tmp.h != opX.h || tmp.l != opX.l) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void STWI(void) {
  /* store wyde immediate */
  Octa tmp;

  mmuWriteWyde(sumYZ, (Wyde) opX.l);
  tmp = sarOcta(shlOcta(opX, 48), 48);
  if (tmp.h != opX.h || tmp.l != opX.l) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void STWU(void) {
  /* store wyde unsigned */
  mmuWriteWyde(sumYZ, (Wyde) opX.l);
}


static void STWUI(void) {
  /* store wyde unsigned immediate */
  mmuWriteWyde(sumYZ, (Wyde) opX.l);
}


static void STT(void) {
  /* store tetra */
  Octa tmp;

  mmuWriteTetra(sumYZ, opX.l);
  tmp = sarOcta(shlOcta(opX, 32), 32);
  if (tmp.h != opX.h || tmp.l != opX.l) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void STTI(void) {
  /* store tetra immediate */
  Octa tmp;

  mmuWriteTetra(sumYZ, opX.l);
  tmp = sarOcta(shlOcta(opX, 32), 32);
  if (tmp.h != opX.h || tmp.l != opX.l) {
    /* integer overflow */
    cpuSetTrip(TRIP_V_BIT);
  }
}


static void STTU(void) {
  /* store tetra unsigned */
  mmuWriteTetra(sumYZ, opX.l);
}


static void STTUI(void) {
  /* store tetra unsigned immediate */
  mmuWriteTetra(sumYZ, opX.l);
}


static void STO(void) {
  /* store octa */
  mmuWriteOcta(sumYZ, opX);
}


static void STOI(void) {
  /* store octa immediate */
  mmuWriteOcta(sumYZ, opX);
}


static void STOU(void) {
  /* store octa unsigned */
  mmuWriteOcta(sumYZ, opX);
}


static void STOUI(void) {
  /* store octa unsigned immediate */
  mmuWriteOcta(sumYZ, opX);
}


/*------------------------------------------------------------*/

/* B0 - BF */


static void STHT(void) {
  /* store high tetra */
  mmuWriteTetra(sumYZ, opX.h);
}


static void STHTI(void) {
  /* store high tetra immediate */
  mmuWriteTetra(sumYZ, opX.h);
}


static void STCO(void) {
  /* store constant octa */
  opX.h = 0;
  opX.l = ZEXT(x, 8);
  mmuWriteOcta(sumYZ, opX);
}


static void STCOI(void) {
  /* store constant octa immediate */
  opX.h = 0;
  opX.l = ZEXT(x, 8);
  mmuWriteOcta(sumYZ, opX);
}


static void STUNC(void) {
  /* store octa uncached */
  mmuWriteOcta(sumYZ, opX);
}


static void STUNCI(void) {
  /* store octa uncached immediate */
  mmuWriteOcta(sumYZ, opX);
}


static void SYNCD(void) {
  /* synchronize data */
  /* has no effect because caches are not simulated */
}


static void SYNCDI(void) {
  /* synchronize data immediate */
  /* has no effect because caches are not simulated */
}


static void PREST(void) {
  /* prestore data */
  /* has no effect because caches are not simulated */
}


static void PRESTI(void) {
  /* prestore data immediate */
  /* has no effect because caches are not simulated */
}


static void SYNCID(void) {
  /* synchronize instructions and data */
  /* has no effect because caches are not simulated */
}


static void SYNCIDI(void) {
  /* synchronize instructions and data immediate */
  /* has no effect because caches are not simulated */
}


static void PUSHGO(void) {
  /* push registers and go */
  push(x);
  gr[rJ] = next;
  next = sumYZ;
}


static void PUSHGOI(void) {
  /* push registers and go immediate */
  push(x);
  gr[rJ] = next;
  next = sumYZ;
}


/*------------------------------------------------------------*/

/* C0 - CF */


static void OR(void) {
  /* bitwise or */
  res = orOcta(opY, opZ);
  *resPtr = res;
}


static void ORI(void) {
  /* bitwise or immediate */
  res = orOcta(opY, opZ);
  *resPtr = res;
}


static void ORN(void) {
  /* bitwise or-not */
  res = ornOcta(opY, opZ);
  *resPtr = res;
}


static void ORNI(void) {
  /* bitwise or-not immediate */
  res = ornOcta(opY, opZ);
  *resPtr = res;
}


static void NOR(void) {
  /* bitwise not-or */
  res = norOcta(opY, opZ);
  *resPtr = res;
}


static void NORI(void) {
  /* bitwise not-or immediate */
  res = norOcta(opY, opZ);
  *resPtr = res;
}


static void XOR(void) {
  /* bitwise exclusive-or */
  res = xorOcta(opY, opZ);
  *resPtr = res;
}


static void XORI(void) {
  /* bitwise exclusive-or immediate */
  res = xorOcta(opY, opZ);
  *resPtr = res;
}


static void AND(void) {
  /* bitwise and */
  res = andOcta(opY, opZ);
  *resPtr = res;
}


static void ANDI(void) {
  /* bitwise and immediate */
  res = andOcta(opY, opZ);
  *resPtr = res;
}


static void ANDN(void) {
  /* bitwise and-not */
  res = andnOcta(opY, opZ);
  *resPtr = res;
}


static void ANDNI(void) {
  /* bitwise and-not immediate */
  res = andnOcta(opY, opZ);
  *resPtr = res;
}


static void NAND(void) {
  /* bitwise not-and */
  res = nandOcta(opY, opZ);
  *resPtr = res;
}


static void NANDI(void) {
  /* bitwise not-and immediate */
  res = nandOcta(opY, opZ);
  *resPtr = res;
}


static void NXOR(void) {
  /* bitwise not-exclusive-or */
  res = nxorOcta(opY, opZ);
  *resPtr = res;
}


static void NXORI(void) {
  /* bitwise not-exclusive-or immediate */
  res = nxorOcta(opY, opZ);
  *resPtr = res;
}


/*------------------------------------------------------------*/

/* D0 - DF */


static void BDIF(void) {
  /* byte difference */
  res = bdifOcta(opY, opZ);
  *resPtr = res;
}


static void BDIFI(void) {
  /* byte difference immediate */
  res = bdifOcta(opY, opZ);
  *resPtr = res;
}


static void WDIF(void) {
  /* wyde difference */
  res = wdifOcta(opY, opZ);
  *resPtr = res;
}


static void WDIFI(void) {
  /* wyde difference immediate */
  res = wdifOcta(opY, opZ);
  *resPtr = res;
}


static void TDIF(void) {
  /* tetra difference */
  res = tdifOcta(opY, opZ);
  *resPtr = res;
}


static void TDIFI(void) {
  /* tetra difference immediate */
  res = tdifOcta(opY, opZ);
  *resPtr = res;
}


static void ODIF(void) {
  /* octa difference */
  res = odifOcta(opY, opZ);
  *resPtr = res;
}


static void ODIFI(void) {
  /* octa difference immediate */
  res = odifOcta(opY, opZ);
  *resPtr = res;
}


static void MUX(void) {
  /* bitwise multiplex */
  res = muxOcta(opY, opZ, gr[rM]);
  *resPtr = res;
}


static void MUXI(void) {
  /* bitwise multiplex immediate */
  res = muxOcta(opY, opZ, gr[rM]);
  *resPtr = res;
}


static void SADD(void) {
  /* sideways add */
  res = saddOcta(opY, opZ);
  *resPtr = res;
}


static void SADDI(void) {
  /* sideways add immediate */
  res = saddOcta(opY, opZ);
  *resPtr = res;
}


static void MOR(void) {
  /* multiple or */
  res = morOcta(opY, opZ);
  *resPtr = res;
}


static void MORI(void) {
  /* multiple or immediate */
  res = morOcta(opY, opZ);
  *resPtr = res;
}


static void MXOR(void) {
  /* multiple exclusive-or */
  res = mxorOcta(opY, opZ);
  *resPtr = res;
}


static void MXORI(void) {
  /* multiple exclusive-or immediate */
  res = mxorOcta(opY, opZ);
  *resPtr = res;
}


/*------------------------------------------------------------*/

/* E0 - EF */


static void SETH(void) {
  /* set to high wyde */
  res = opZ;
  *resPtr = res;
}


static void SETMH(void) {
  /* set to medium high wyde */
  res = opZ;
  *resPtr = res;
}


static void SETML(void) {
  /* set to medium low wyde */
  res = opZ;
  *resPtr = res;
}


static void SETL(void) {
  /* set to low wyde */
  res = opZ;
  *resPtr = res;
}


static void INCH(void) {
  /* increase by high wyde */
  res = addOcta(opX, opZ);
  *resPtr = res;
}


static void INCMH(void) {
  /* increase by medium high wyde */
  res = addOcta(opX, opZ);
  *resPtr = res;
}


static void INCML(void) {
  /* increase by medium low wyde */
  res = addOcta(opX, opZ);
  *resPtr = res;
}


static void INCL(void) {
  /* increase by low wyde */
  res = addOcta(opX, opZ);
  *resPtr = res;
}


static void ORH(void) {
  /* bitwise or with high wyde */
  res = orOcta(opX, opZ);
  *resPtr = res;
}


static void ORMH(void) {
  /* bitwise or with medium high wyde */
  res = orOcta(opX, opZ);
  *resPtr = res;
}


static void ORML(void) {
  /* bitwise or with medium low wyde */
  res = orOcta(opX, opZ);
  *resPtr = res;
}


static void ORL(void) {
  /* bitwise or with low wyde */
  res = orOcta(opX, opZ);
  *resPtr = res;
}


static void ANDNH(void) {
  /* bitwise and-not high wyde */
  res = andnOcta(opX, opZ);
  *resPtr = res;
}


static void ANDNMH(void) {
  /* bitwise and-not medium high wyde */
  res = andnOcta(opX, opZ);
  *resPtr = res;
}


static void ANDNML(void) {
  /* bitwise and-not medium low wyde */
  res = andnOcta(opX, opZ);
  *resPtr = res;
}


static void ANDNL(void) {
  /* bitwise and-not low wyde */
  res = andnOcta(opX, opZ);
  *resPtr = res;
}


/*------------------------------------------------------------*/

/* F0 - FF */


static void JMP(void) {
  /* jump */
  next = altNext;
}


static void JMPB(void) {
  /* jump back */
  next = altNext;
}


static void PUSHJ(void) {
  /* push registers and jump */
  push(x);
  gr[rJ] = next;
  next = altNext;
}


static void PUSHJB(void) {
  /* push registers and jump back */
  push(x);
  gr[rJ] = next;
  next = altNext;
}


static void GETA(void) {
  /* get address */
  res = altNext;
  *resPtr = res;
}


static void GETAB(void) {
  /* get address back */
  res = altNext;
  *resPtr = res;
}


static void PUT(void) {
  /* put into special register */
  if (y != 0 || x >= 32) {
    throwException(EXC_DT_B);
  }
  writeSpecial(x, opZ);
}


static void PUTI(void) {
  /* put into special register immediate */
  if (y != 0 || x >= 32) {
    throwException(EXC_DT_B);
  }
  writeSpecial(x, opZ);
}


static void POP(void) {
  /* pop registers and return from subroutine */
  pop(x);
  opY = gr[rJ];
  opZ.h = 0;
  opZ.l = LIT16(instr) << 2;
  next = addOcta(opY, opZ);
}


static void RESUME(void) {
  /* resume after interrupt */
  if (x != 0 || y != 0 || z > 1) {
    throwException(EXC_DT_B);
  }
  if (z == 0) {
    if (debugInstr) {
      cPrintf("INSTR: RESUME 0 encountered\n");
    }
    if ((gr[rX].h & SIGN(32)) != 0) {
      /* rX is negative: ignore instruction in lower half of rX */
      if (debugInstr) {
        cPrintf("INSTR: instruction in lower half of rX ignored\n");
      }
      next = gr[rW];
      resuming = RESUME_NONE;
    } else {
      /* rX is non-negative: insert instruction in lower half of rX */
      if (debugInstr) {
        cPrintf("INSTR: instruction in lower half of rX inserted\n");
      }
      ropcode = (gr[rX].h >> 24) & 0xFF;
      rexcept = gr[rX].h & 0x0000FF00;
      if (ropcode > ROP_SET) {
        throwException(EXC_DT_B);
      }
      next = incOcta(gr[rW], -4);
      instr = gr[rX].l;
      resuming = RESUME_0;
    }
  } else
  if (z == 1) {
    if (debugInstr) {
      cPrintf("INSTR: RESUME 1 encountered\n");
    }
    if ((ip.h & SIGN(32)) == 0) {
      throwException(EXC_DT_K);
    }
    if ((gr[rXX].h & SIGN(32)) != 0) {
      /* rXX is negative: ignore instruction in lower half of rXX */
      if (debugInstr) {
        cPrintf("INSTR: instruction in lower half of rXX ignored\n");
      }
      next = gr[rWW];
      resuming = RESUME_NONE;
    } else {
      /* rXX is non-negative: insert instruction in lower half of rXX */
      if (debugInstr) {
        cPrintf("INSTR: instruction in lower half of rXX inserted\n");
      }
      ropcode = (gr[rXX].h >> 24) & 0xFF;
      rexcept = gr[rXX].h & 0x0000FF00;
      if (ropcode > ROP_MAP) {
        throwException(EXC_DT_B);
      }
      next = incOcta(gr[rWW], -4);
      instr = gr[rXX].l;
      resuming = RESUME_1;
    }
    gr[rK] = gr[255];
    gr[255] = gr[rBB];
  }
}


static void SYNC(void) {
  /* synchronize */
  if (x != 0 || y != 0 || z > 7) {
    throwException(EXC_DT_B);
  }
  if (z > 3 && (ip.h & SIGN(32)) == 0) {
    throwException(EXC_DT_K);
  }
  switch (z) {
    case 0:
      /* has no effect because pipeline is not simulated */
      break;
    case 1:
      /* has no effect because pipeline is not simulated */
      break;
    case 2:
      /* has no effect because pipeline is not simulated */
      break;
    case 3:
      /* has no effect because pipeline is not simulated */
      break;
    case 4:
      /* has no effect because power-saver mode is not simulated */
      break;
    case 5:
      /* has no effect because caches are not simulated */
      break;
    case 6:
      /* clear virtual address translation caches */
      mmuRemoveAll(ITC);
      mmuRemoveAll(DTC);
      break;
    case 7:
      /* has no effect because caches are not simulated */
      break;
  }
}


static void GET(void) {
  /* get from special register */
  if (y != 0 || z >= 32) {
    throwException(EXC_DT_B);
  }
  res = readSpecial(z);
  *resPtr = res;
}


static void TRIP(void) {
  /* trip */
  cpuSetTrip(TRIP_H_BIT);
}


/*------------------------------------------------------------*/


#define PUSH_POP_BIT		0x80
#define REL_ADDR_BIT		0x40
#define X_IS_DST_BIT		0x20
#define X_IS_SRC_BIT		0x10
#define Y_IS_SRC_BIT		0x08
#define Y_IS_IMM_BIT		0x04
#define Z_IS_SRC_BIT		0x02
#define Z_IS_IMM_BIT		0x01


static Byte instrFlags[256] = {
  /* 00 */  0x0A, 0x2A, 0x2A, 0x2A, 0x2A, 0x26, 0x2A, 0x26,
  /* 08 */  0x26, 0x25, 0x26, 0x25, 0x26, 0x25, 0x26, 0x25,
  /* 10 */  0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x26, 0x2A, 0x26,
  /* 18 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* 20 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* 28 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* 30 */  0x2A, 0x29, 0x2A, 0x29, 0x26, 0x25, 0x26, 0x25,
  /* 38 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* 40 */  0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50,
  /* 48 */  0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50,
  /* 50 */  0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50,
  /* 58 */  0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50,
  /* 60 */  0x3A, 0x39, 0x3A, 0x39, 0x3A, 0x39, 0x3A, 0x39,
  /* 68 */  0x3A, 0x39, 0x3A, 0x39, 0x3A, 0x39, 0x3A, 0x39,
  /* 70 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* 78 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* 80 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* 88 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* 90 */  0x2A, 0x29, 0x2A, 0x29, 0x3A, 0x39, 0x2A, 0x29,
  /* 98 */  0x2A, 0x29, 0x0A, 0x09, 0x0A, 0x09, 0x2A, 0x29,
  /* A0 */  0x1A, 0x19, 0x1A, 0x19, 0x1A, 0x19, 0x1A, 0x19,
  /* A8 */  0x1A, 0x19, 0x1A, 0x19, 0x1A, 0x19, 0x1A, 0x19,
  /* B0 */  0x1A, 0x19, 0x1A, 0x19, 0x0A, 0x09, 0x1A, 0x19,
  /* B8 */  0x0A, 0x09, 0x0A, 0x09, 0x0A, 0x09, 0xAA, 0xA9,
  /* C0 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* C8 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* D0 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* D8 */  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,
  /* E0 */  0x20, 0x20, 0x20, 0x20, 0x30, 0x30, 0x30, 0x30,
  /* E8 */  0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
  /* F0 */  0x40, 0x40, 0xE0, 0xE0, 0x60, 0x60, 0x02, 0x01,
  /* F8 */  0x80, 0x00, 0x20, 0x82, 0x01, 0x00, 0x20, 0x0A
};


static void (*instrAction[256])(void) = {
  /* 00 */  TRAP,   emu,    emu,    emu,    emu,    emu,    emu,    emu,
  /* 08 */  emu,    emu,    emu,    emu,    emu,    emu,    emu,    emu,
  /* 10 */  emu,    emu,    emu,    emu,    emu,    emu,    emu,    emu,
  /* 18 */  MUL,    MULI,   MULU,   MULUI,  DIV,    DIVI,   DIVU,   DIVUI,
  /* 20 */  ADD,    ADDI,   ADDU,   ADDUI,  SUB,    SUBI,   SUBU,   SUBUI,
  /* 28 */  ADDU2,  ADDUI2, ADDU4,  ADDUI4, ADDU8,  ADDUI8, ADDU16, ADDUI16,
  /* 30 */  CMP,    CMPI,   CMPU,   CMPUI,  NEG,    NEGI,   NEGU,   NEGUI,
  /* 38 */  SL,     SLI,    SLU,    SLUI,   SR,     SRI,    SRU,    SRUI,
  /* 40 */  BN,     BNB,    BZ,     BZB,    BP,     BPB,    BOD,    BODB,
  /* 48 */  BNN,    BNNB,   BNZ,    BNZB,   BNP,    BNPB,   BEV,    BEVB,
  /* 50 */  PBN,    PBNB,   PBZ,    PBZB,   PBP,    PBPB,   PBOD,   PBODB,
  /* 58 */  PBNN,   PBNNB,  PBNZ,   PBNZB,  PBNP,   PBNPB,  PBEV,   PBEVB,
  /* 60 */  CSN,    CSNI,   CSZ,    CSZI,   CSP,    CSPI,   CSOD,   CSODI,
  /* 68 */  CSNN,   CSNNI,  CSNZ,   CSNZI,  CSNP,   CSNPI,  CSEV,   CSEVI,
  /* 70 */  ZSN,    ZSNI,   ZSZ,    ZSZI,   ZSP,    ZSPI,   ZSOD,   ZSODI,
  /* 78 */  ZSNN,   ZSNNI,  ZSNZ,   ZSNZI,  ZSNP,   ZSNPI,  ZSEV,   ZSEVI,
  /* 80 */  LDB,    LDBI,   LDBU,   LDBUI,  LDW,    LDWI,   LDWU,   LDWUI,
  /* 88 */  LDT,    LDTI,   LDTU,   LDTUI,  LDO,    LDOI,   LDOU,   LDOUI,
  /* 90 */  emu,    emu,    LDHT,   LDHTI,  CSWAP,  CSWAPI, LDUNC,  LDUNCI,
  /* 98 */  LDVTS,  LDVTSI, PRELD,  PRELDI, PREGO,  PREGOI, GO,     GOI,
  /* A0 */  STB,    STBI,   STBU,   STBUI,  STW,    STWI,   STWU,   STWUI,
  /* A8 */  STT,    STTI,   STTU,   STTUI,  STO,    STOI,   STOU,   STOUI,
  /* B0 */  emu,    emu,    STHT,   STHTI,  STCO,   STCOI,  STUNC,  STUNCI,
  /* B8 */  SYNCD,  SYNCDI, PREST,  PRESTI, SYNCID, SYNCIDI,PUSHGO, PUSHGOI,
  /* C0 */  OR,     ORI,    ORN,    ORNI,   NOR,    NORI,   XOR,    XORI,
  /* C8 */  AND,    ANDI,   ANDN,   ANDNI,  NAND,   NANDI,  NXOR,   NXORI,
  /* D0 */  BDIF,   BDIFI,  WDIF,   WDIFI,  TDIF,   TDIFI,  ODIF,   ODIFI,
  /* D8 */  MUX,    MUXI,   SADD,   SADDI,  MOR,    MORI,   MXOR,   MXORI,
  /* E0 */  SETH,   SETMH,  SETML,  SETL,   INCH,   INCMH,  INCML,  INCL,
  /* E8 */  ORH,    ORMH,   ORML,   ORL,    ANDNH,  ANDNMH, ANDNML, ANDNL,
  /* F0 */  JMP,    JMPB,   PUSHJ,  PUSHJB, GETA,   GETAB,  PUT,    PUTI,
  /* F8 */  POP,    RESUME, emu,    emu,    SYNC,   emu,    GET,    TRIP
};


static void execNextInstruction(void) {
  int offset;
  Tetra trips;
  int i;

  /* the RESUME instruction requires a loop here */
  resuming = RESUME_NONE;
  do {
    /* get instruction */
    if (resuming == RESUME_NONE) {
      /* fetch instruction from memory */
      if (debugInstr) {
        cPrintf("INSTR: ** normal fetch **\n");
      }
      instr = mmuGetInstr(ip);
    } else {
      /* instruction has already been fetched from execution register */
      if (debugInstr) {
        cPrintf("INSTR: ** resuming **\n");
      }
    }
    next = incOcta(ip, 4);
    /* decode instruction */
    opcode = OPCODE(instr);
    x = DST(instr);
    y = SRC1(instr);
    z = SRC2(instr);
    opflags = instrFlags[opcode];
    /* compute alternative 'next' value for jumps */
    if (opflags & REL_ADDR_BIT) {
      if ((opcode & 0xFE) == 0xF0) {
        /* 3 byte offset */
        offset = LIT24(instr);
        if (opcode & 0x01) {
          /* back */
          offset -= 0x1000000;
        }
      } else {
        /* 2 byte offset */
        offset = LIT16(instr);
        if (opcode & 0x01) {
          /* back */
          offset -= 0x10000;
        }
      }
      altNext = incOcta(ip, offset << 2);
    }
    /* install operands */
    if (resuming == RESUME_NONE || ropcode == ROP_AGAIN) {
      /* use operand specification from actual instruction */
      if (opflags & X_IS_SRC_BIT) {
        /* set opX from register x */
        opX = readDR(x);
      } else {
        /* set opX to 0 */
        opX = zero;
      }
      if (opflags & Z_IS_IMM_BIT) {
        /* set opZ from immediate value z */
        opZ.h = 0;
        opZ.l = z;
      } else
      if (opflags & Z_IS_SRC_BIT) {
        /* set opZ from register z */
        opZ = readDR(z);
      } else
      if ((opcode & 0xF0) == 0xE0) {
        /* set opZ as an immediate wyde */
        switch (opcode & 3) {
          case 0:
            opZ.h = LIT16(instr) << 16;
            opZ.l = 0;
            break;
          case 1:
            opZ.h = LIT16(instr);
            opZ.l = 0;
            break;
          case 2:
            opZ.h = 0;
            opZ.l = LIT16(instr) << 16;
            break;
          case 3:
            opZ.h = 0;
            opZ.l = LIT16(instr);
            break;
        }
      } else {
        /* set opZ to 0 */
        opZ = zero;
      }
      if (opflags & Y_IS_IMM_BIT) {
        /* set opY from immediate value y */
        opY.h = 0;
        opY.l = y;
      } else
      if (opflags & Y_IS_SRC_BIT) {
        /* set opY from register y */
        opY = readDR(y);
      } else {
        /* set opY to 0 */
        opY = zero;
      }
    } else {
      /* use operands from operand registers */
      if (resuming == RESUME_0) {
        if (ropcode == ROP_CONT) {
          opY = gr[rY];
          opZ = gr[rZ];
        } else
        if (ropcode == ROP_SET) {
          opcode = 0xC1;
          opY = gr[rZ];
          opZ = zero;
          opflags = X_IS_DST_BIT;
        }
      } else
      if (resuming == RESUME_1) {
        if (ropcode == ROP_CONT) {
          opY = gr[rYY];
          opZ = gr[rZZ];
        } else
        if (ropcode == ROP_SET) {
          opcode = 0xC1;
          opY = gr[rZZ];
          opZ = zero;
          opflags = X_IS_DST_BIT;
        } else
        if (ropcode == ROP_MAP) {
        }
      }
    }
    if (opflags & X_IS_DST_BIT) {
      /* install register x as the destination */
      resPtr = writeDR(x);
    }
    sumYZ = addOcta(opY, opZ);
    /* output debug information if requested */
    if (debugInstr) {
      cPrintf("INSTR: ip = #%08X%08X, instr = #%08X",
              ip.h, ip.l, instr);
      if (instrCodeTbl[opcode] == NULL) {
        cPrintf(" (%s)\n", "???");
      } else {
        cPrintf(" (%s)\n", instrCodeTbl[opcode]->name);
      }
      cPrintf("INSTR: opcode = #%02X, x = #%02X", opcode, x);
      cPrintf(", y = #%02X, z = #%02X, flags = #%02X\n", y, z, opflags);
      if (opflags & PUSH_POP_BIT) {
        cPrintf("INSTR: push/pop instruction\n");
      }
      if (opflags & REL_ADDR_BIT) {
        cPrintf("INSTR: rel addr --> addr = #%08X%08X\n",
                altNext.h, altNext.l);
      }
      if (opflags & X_IS_DST_BIT) {
        cPrintf("INSTR: x is dst --> store to $%d\n",
                x);
      }
      if (opflags & X_IS_SRC_BIT) {
        cPrintf("INSTR: x is src --> read from $%d, opX = #%08X%08X\n",
                x, opX.h, opX.l);
      }
      if (opflags & Y_IS_SRC_BIT) {
        cPrintf("INSTR: y is src --> read from $%d, opY = #%08X%08X\n",
                y, opY.h, opY.l);
      }
      if (opflags & Y_IS_IMM_BIT) {
        cPrintf("INSTR: y is imm --> opY = #%08X%08X\n",
                opY.h, opY.l);
      }
      if (opflags & Z_IS_SRC_BIT) {
        cPrintf("INSTR: z is src --> read from $%d, opZ = #%08X%08X\n",
                z, opZ.h, opZ.l);
      }
      if (opflags & Z_IS_IMM_BIT) {
        cPrintf("INSTR: z is imm --> opZ = #%08X%08X\n",
                opZ.h, opZ.l);
      }
    }
    /* execute instruction */
    resuming = RESUME_NONE;
    tripsPending = 0;
    (*instrAction[opcode])();
    /* check for trip interrupt */
    if ((tripsPending & (TRIP_U_BIT | TRIP_X_BIT)) == TRIP_U_BIT &&
        (gr[rA].l & TRIP_U_BIT) == 0) {
      /* ignore underflow that is exact and not enabled */
      tripsPending &= ~TRIP_U_BIT;
    }
    if (tripsPending != 0) {
      /* instruction tripped */
      if (debugTrips) {
        cPrintf("TRIP: instruction tripped, tripsPending = #%08X\n",
                tripsPending);
      }
      trips = tripsPending & (TRIP_H_BIT | (gr[rA].l & 0x0000FF00));
      if (trips != 0 && (ip.h & SIGN(32)) == 0) {
        /* initiate a trip interrupt */
        for (i = 0; (trips & TRIP_H_BIT) == 0; i++) {
          trips <<= 1;
        }
        tripsPending &= ~(TRIP_H_BIT >> i);
        gr[rW] = next;
        next.h = 0;
        next.l = i << 4;
        gr[rX].h = 0x80000000;
        gr[rX].l = instr;
        if ((opcode & 0xE0) == 0xA0) {
          /* this is a store instruction */
          gr[rY] = sumYZ;
          gr[rZ] = opX;
        } else {
          /* this is any other instruction */
          gr[rY] = opY;
          gr[rZ] = opZ;
        }
        gr[rB] = gr[255];
        gr[255] = gr[rJ];
        if (debugTrips) {
          cPrintf("TRIP: trip interrupt initiated for trip %d (%c)\n",
                  i, "HDVWIOUZX"[i]);
        }
      } else {
        /* do not initiate a trip interrupt */
        if (debugTrips) {
          cPrintf("TRIP: no trip interrupt initiated ");
          cPrintf("(masked or from negative location)\n");
        }
      }
      gr[rA].l |= (tripsPending & ~TRIP_H_BIT) >> 8;
    }
    /* update ip */
    ip = next;
  } while (resuming != RESUME_NONE);
}


/**************************************************************/


Octa cpuGetIP(void) {
  return ip;
}


void cpuSetIP(Octa addr) {
  addr.l &= ~0x00000003;
  ip = addr;
}


Octa cpuGetGR(int regnum) {
  regnum &= GR_MASK;
  if (regnum < 32) {
    return readSpecial(regnum);
  } else {
    return gr[regnum];
  }
}


void cpuSetGR(int regnum, Octa value) {
  regnum &= GR_MASK;
  if (regnum < 32) {
    writeSpecial(regnum, value);
  } else {
    gr[regnum] = value;
  }
}


Octa cpuGetLR(int regnum) {
  regnum &= LR_MASK;
  return lr[regnum];
}


void cpuSetLR(int regnum, Octa value) {
  regnum &= LR_MASK;
  lr[regnum] = value;
}


Octa cpuGetDR(int regnum) {
  regnum &= 0xFF;
  return readDR(regnum);
}


void cpuSetDR(int regnum, Octa value) {
  regnum &= 0xFF;
  *writeDR(regnum) = value;
}


Bool cpuTestBreak(void) {
  return breakSet;
}


Octa cpuGetBreak(void) {
  return breakAddr;
}


void cpuSetBreak(Octa addr) {
  addr.l &= ~0x00000003;
  breakAddr = addr;
  breakSet = TRUE;
}


void cpuResetBreak(void) {
  breakSet = FALSE;
}


void cpuStep(void) {
  jmp_buf myEnvironment;
  int exception;

  exception = setjmp(myEnvironment);
  if (exception == EXC_NONE) {
    /* initialization */
    pushEnvironment(&myEnvironment);
    handleRealTimeTasks();
    execNextInstruction();
    handleTraps();
  } else {
    /* an exception was thrown */
    cpuSetTrap(exception);
    handleTraps();
  }
  popEnvironment();
}


void cpuRun(void) {
  jmp_buf myEnvironment;
  int exception;

  run = TRUE;
  exception = setjmp(myEnvironment);
  if (exception == EXC_NONE) {
    /* initialization */
    pushEnvironment(&myEnvironment);
  } else {
    /* an exception was thrown */
    cpuSetTrap(exception);
    handleTraps();
    if (breakSet && ip.h == breakAddr.h && ip.l == breakAddr.l) {
      run = FALSE;
    }
  }
  while (run) {
    handleRealTimeTasks();
    execNextInstruction();
    handleTraps();
    if (breakSet && ip.h == breakAddr.h && ip.l == breakAddr.l) {
      run = FALSE;
    }
  }
  popEnvironment();
}


void cpuHalt(void) {
  run = FALSE;
}


void cpuSetTrip(Tetra tripBitMask) {
  if (debugTrips) {
    cPrintf("**** TRIP set: bit mask = #%08X ****\n", tripBitMask);
  }
  tripsPending |= tripBitMask;
}


void cpuSetTrap(int trapBitNumber) {
  if (debugTraps) {
    cPrintf("**** TRAP set: bit number = %d ****\n", trapBitNumber);
  }
  if (trapBitNumber < 0) {
    trapBitNumber = -trapBitNumber;
    trapBitNumber--;
    ftrapsPending |= (1 << trapBitNumber);
  } else {
    trapBitNumber--;
    if (trapBitNumber >= 32) {
      dtrapsPending.h |= (1 << (trapBitNumber - 32));
    } else {
      dtrapsPending.l |= (1 << trapBitNumber);
    }
  }
}


void cpuReset(void) {
  int i;

  cPrintf("Resetting CPU...\n");
  /* preset registers */
  ip.h = SIGN(32) | ROM_BASE_H;
  ip.l = ROM_BASE_L;
  for (i = 0; i < GR_NUM; i++) {
    gr[i] = zero;
  }
  for (i = 0; i < LR_NUM; i++) {
    lr[i] = zero;
  }
  G = 255;
  gr[rG].h = 0;
  gr[rG].l = G;
  L = 0;
  gr[rL].h = 0;
  gr[rL].l = L;
  O = 0;
  gr[rO].h = 0;
  gr[rO].l = O;
  S = 0;
  gr[rS].h = 0;
  gr[rS].l = S;
  gr[rN].h = (VERSION << 24) | (SUBVERSION << 16) | (SUBSUBVERSION <<8);
  gr[rN].l = ABSTIME;
  /* reset simulator control variables */
  breakSet = FALSE;
  instrCount = 0;
  ftrapsPending = 0;
  dtrapsPending = zero;
}


void cpuInit(void) {
  cpuReset();
}


void cpuExit(void) {
}
