/*
 * memory.c -- main memory simulation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "except.h"
#include "arith.h"
#include "memory.h"
#include "term.h"
#include "disk.h"
#include "graph.h"


#define IMAGE_BUFSIZE		100


static U32 romSize;
static Byte *rom;

static U32 memSize;
static Byte *mem;

static FILE *romImage;
static FILE *progImage;


Octa memoryReadOcta(Octa pAddr) {
  Octa data;

  pAddr.l &= ~0x00000007;
  if (pAddr.h == MEM_BASE_H && pAddr.l < memSize) {
    /* memory */
    data.h = ((Tetra) *(mem + pAddr.l + 0)) << 24 |
             ((Tetra) *(mem + pAddr.l + 1)) << 16 |
             ((Tetra) *(mem + pAddr.l + 2)) <<  8 |
             ((Tetra) *(mem + pAddr.l + 3)) <<  0;
    data.l = ((Tetra) *(mem + pAddr.l + 4)) << 24 |
             ((Tetra) *(mem + pAddr.l + 5)) << 16 |
             ((Tetra) *(mem + pAddr.l + 6)) <<  8 |
             ((Tetra) *(mem + pAddr.l + 7)) <<  0;
    return data;
  }
  if (pAddr.h == ROM_BASE_H && pAddr.l < romSize) {
    /* ROM */
    data.h = ((Tetra) *(rom + pAddr.l + 0)) << 24 |
             ((Tetra) *(rom + pAddr.l + 1)) << 16 |
             ((Tetra) *(rom + pAddr.l + 2)) <<  8 |
             ((Tetra) *(rom + pAddr.l + 3)) <<  0;
    data.l = ((Tetra) *(rom + pAddr.l + 4)) << 24 |
             ((Tetra) *(rom + pAddr.l + 5)) << 16 |
             ((Tetra) *(rom + pAddr.l + 6)) <<  8 |
             ((Tetra) *(rom + pAddr.l + 7)) <<  0;
    return data;
  }
  if ((pAddr.h & DEV_TYPE_MASK) == TERM_BASE_H) {
    /* terminal controller */
    data = termRead(pAddr.h & DEV_NUM_MASK, pAddr.l);
    return data;
  }
  if ((pAddr.h & DEV_TYPE_MASK) == DISK_BASE_H) {
    /* disk controller */
    data = diskRead(pAddr.h & DEV_NUM_MASK, pAddr.l);
    return data;
  }
  if ((pAddr.h & DEV_TYPE_MASK) == GRAPH_BASE_H) {
    /* graphics controller */
    data = graphRead(pAddr.h & DEV_NUM_MASK, pAddr.l);
    return data;
  }
  /* throw bus exception */
  throwException(EXC_DT_MEMERR);
  /* not reached */
  data.h = 0;
  data.l = 0;
  return data;
}


Tetra memoryReadTetra(Octa pAddr) {
  Tetra data;

  pAddr.l &= ~0x00000003;
  if (pAddr.h == MEM_BASE_H && pAddr.l < memSize) {
    /* memory */
    data = ((Tetra) *(mem + pAddr.l + 0)) << 24 |
           ((Tetra) *(mem + pAddr.l + 1)) << 16 |
           ((Tetra) *(mem + pAddr.l + 2)) <<  8 |
           ((Tetra) *(mem + pAddr.l + 3)) <<  0;
    return data;
  }
  if (pAddr.h == ROM_BASE_H && pAddr.l < romSize) {
    /* ROM */
    data = ((Tetra) *(rom + pAddr.l + 0)) << 24 |
           ((Tetra) *(rom + pAddr.l + 1)) << 16 |
           ((Tetra) *(rom + pAddr.l + 2)) <<  8 |
           ((Tetra) *(rom + pAddr.l + 3)) <<  0;
    return data;
  }
  /* throw bus exception */
  throwException(EXC_DT_MEMERR);
  /* not reached */
  data = 0;
  return data;
}


Wyde memoryReadWyde(Octa pAddr) {
  Wyde data;

  pAddr.l &= ~0x00000001;
  if (pAddr.h == MEM_BASE_H && pAddr.l < memSize) {
    /* memory */
    data = ((Wyde) *(mem + pAddr.l + 0)) << 8 |
           ((Wyde) *(mem + pAddr.l + 1)) << 0;
    return data;
  }
  if (pAddr.h == ROM_BASE_H && pAddr.l < romSize) {
    /* ROM */
    data = ((Wyde) *(rom + pAddr.l + 0)) << 8 |
           ((Wyde) *(rom + pAddr.l + 1)) << 0;
    return data;
  }
  /* throw bus exception */
  throwException(EXC_DT_MEMERR);
  /* not reached */
  data = 0;
  return data;
}


Byte memoryReadByte(Octa pAddr) {
  Byte data;

  if (pAddr.h == MEM_BASE_H && pAddr.l < memSize) {
    /* memory */
    data = ((Byte) *(mem + pAddr.l + 0)) << 0;
    return data;
  }
  if (pAddr.h == ROM_BASE_H && pAddr.l < romSize) {
    /* ROM */
    data = ((Byte) *(rom + pAddr.l + 0)) << 0;
    return data;
  }
  /* throw bus exception */
  throwException(EXC_DT_MEMERR);
  /* not reached */
  data = 0;
  return data;
}


void memoryWriteOcta(Octa pAddr, Octa data) {
  pAddr.l &= ~0x00000007;
  if (pAddr.h == MEM_BASE_H && pAddr.l < memSize) {
    /* memory */
    *(mem + pAddr.l + 0) = (Byte) (data.h >> 24);
    *(mem + pAddr.l + 1) = (Byte) (data.h >> 16);
    *(mem + pAddr.l + 2) = (Byte) (data.h >>  8);
    *(mem + pAddr.l + 3) = (Byte) (data.h >>  0);
    *(mem + pAddr.l + 4) = (Byte) (data.l >> 24);
    *(mem + pAddr.l + 5) = (Byte) (data.l >> 16);
    *(mem + pAddr.l + 6) = (Byte) (data.l >>  8);
    *(mem + pAddr.l + 7) = (Byte) (data.l >>  0);
    return;
  }
  if ((pAddr.h & DEV_TYPE_MASK) == TERM_BASE_H) {
    /* terminal controller */
    termWrite(pAddr.h & DEV_NUM_MASK, pAddr.l, data);
    return;
  }
  if ((pAddr.h & DEV_TYPE_MASK) == DISK_BASE_H) {
    /* disk controller */
    diskWrite(pAddr.h & DEV_NUM_MASK, pAddr.l, data);
    return;
  }
  if ((pAddr.h & DEV_TYPE_MASK) == GRAPH_BASE_H) {
    /* graphics controller */
    graphWrite(pAddr.h & DEV_NUM_MASK, pAddr.l, data);
    return;
  }
  /* throw bus exception */
  throwException(EXC_DT_MEMERR);
}


void memoryWriteTetra(Octa pAddr, Tetra data) {
  pAddr.l &= ~0x00000003;
  if (pAddr.h == MEM_BASE_H && pAddr.l < memSize) {
    /* memory */
    *(mem + pAddr.l + 0) = (Byte) (data >> 24);
    *(mem + pAddr.l + 1) = (Byte) (data >> 16);
    *(mem + pAddr.l + 2) = (Byte) (data >>  8);
    *(mem + pAddr.l + 3) = (Byte) (data >>  0);
    return;
  }
  /* throw bus exception */
  throwException(EXC_DT_MEMERR);
}


void memoryWriteWyde(Octa pAddr, Wyde data) {
  pAddr.l &= ~0x00000001;
  if (pAddr.h == MEM_BASE_H && pAddr.l < memSize) {
    /* memory */
    *(mem + pAddr.l + 0) = (Byte) (data >> 8);
    *(mem + pAddr.l + 1) = (Byte) (data >> 0);
    return;
  }
  /* throw bus exception */
  throwException(EXC_DT_MEMERR);
}


void memoryWriteByte(Octa pAddr, Byte data) {
  if (pAddr.h == MEM_BASE_H && pAddr.l < memSize) {
    /* memory */
    *(mem + pAddr.l + 0) = (Byte) (data >> 0);
    return;
  }
  /* throw bus exception */
  throwException(EXC_DT_MEMERR);
}


static U32 loadImage(FILE *image, Octa base, U32 size, Byte *m) {
  char buffer[IMAGE_BUFSIZE];
  U32 count;
  Octa curLoc;
  Octa curDat;
  Octa offset;

  fseek(image, 0, SEEK_SET);
  count = 0;
  curLoc = base;
  while (1) {
    if (fgets(buffer, IMAGE_BUFSIZE, image) == NULL) {
      break;
    }
    if (buffer[strlen(buffer) - 1] != '\n') {
      error("line too long in program or ROM image file");
    }
    if (buffer[12] == ':') {
      /* change the current location */
      if (sscanf(buffer, "%4x%8x", &curLoc.h, &curLoc.l) != 2) {
        error("improper loaction in program or ROM image file");
      }
      curLoc.l &= ~0x00000007;
    } else
    if (buffer[0] == ' ') {
      /* read and store an octabyte */
      if (sscanf(buffer + 1, "%8x%8x", &curDat.h, &curDat.l) != 2) {
        error("improper data in program or ROM image file");
      }
      offset = subOcta(curLoc, base);
      if (offset.h != 0 || offset.l >= size) {
        error("address out of range in program or ROM image file");
      }
      m[offset.l + 0] = (Byte) (curDat.h >> 24);
      m[offset.l + 1] = (Byte) (curDat.h >> 16);
      m[offset.l + 2] = (Byte) (curDat.h >>  8);
      m[offset.l + 3] = (Byte) (curDat.h >>  0);
      m[offset.l + 4] = (Byte) (curDat.l >> 24);
      m[offset.l + 5] = (Byte) (curDat.l >> 16);
      m[offset.l + 6] = (Byte) (curDat.l >>  8);
      m[offset.l + 7] = (Byte) (curDat.l >>  0);
      /* advance current loaction and increment load count */
      curLoc = incOcta(curLoc, 8);
      count += 8;
    } else {
      /* neither address nor data */
      error("improper program or ROM image file");
    }
  }
  return count;
}


void memoryReset(void) {
  U32 i;
  Octa base;

  cPrintf("Resetting Memory...\n");
  for (i = 0; i < romSize; i++) {
    rom[i] = rand();
  }
  for (i = 0; i < memSize; i++) {
    mem[i] = rand();
  }
  if (romImage != NULL) {
    for (i = 0; i < romSize; i++) {
      rom[i] = 0;
    }
    base.h = ROM_BASE_H;
    base.l = ROM_BASE_L;
    i = loadImage(romImage, base, romSize, rom);
    cPrintf("ROM of size %u/%u bytes installed.\n", i, romSize);
  }
  if (progImage != NULL) {
    base.h = MEM_BASE_H;
    base.l = MEM_BASE_L;
    i = loadImage(progImage, base, memSize, mem);
    cPrintf("Program of size %u/%u loaded.\n", i, memSize);
  }
}


void memoryInit(U32 memorySize, char *romName, char *progName) {
  /* allocate ROM */
  romSize = ROM_SIZE;
  rom = malloc(romSize);
  if (rom == NULL) {
    error("cannot allocate ROM");
  }
  /* allocate main memory */
  memSize = memorySize;
  mem = malloc(memSize);
  if (mem == NULL) {
    error("cannot allocate main memory");
  }
  /* possibly load ROM image */
  if (romName == NULL) {
    /* no ROM to plug in */
    romImage = NULL;
  } else {
    /* plug in ROM */
    romImage = fopen(romName, "rt");
    if (romImage == NULL) {
      error("cannot open ROM image file '%s'", romName);
    }
  }
  /* possibly load program image */
  if (progName == NULL) {
    /* no program to load */
    progImage = NULL;
  } else {
    /* load program */
    progImage = fopen(progName, "rt");
    if (progImage == NULL) {
      error("cannot open program image file '%s'", progName);
    }
  }
  memoryReset();
}


void memoryExit(void) {
  /* free main memory */
  free(mem);
  /* free ROM */
  free(rom);
}
