/*
 * mmu.c -- MMU simulation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "arith.h"
#include "mmu.h"
#include "memory.h"


#define TC_R		0	/* read */
#define TC_W		1	/* write */
#define TC_X		2	/* execute */


/**************************************************************/


static Bool debugUse = FALSE;
static Bool debugCtrl = TRUE;

static TC_Entry itc[ITC_SIZE];
static TC_Entry dtc[DTC_SIZE];

static Octa currPageMask;	/* current page mask 0xF..F0..0 */
static Octa currOffsetMask;	/* current offset mask 0x0..0F..F */
static int currASN;		/* current address space number */


/**************************************************************/


static TC_Entry *assoc(int asn, Octa page, TC_Entry *tc, int size) {
  int i;

  for (i = 0; i < size; i++) {
    if (tc->valid && tc->asn == asn &&
        tc->page.h == page.h && tc->page.l == page.l) {
      /* key found */
      return tc;
    }
    tc++;
  }
  /* key not found */
  return NULL;
}


static Octa v2p(Octa vAddr, int cacheID, int accessID) {
  Octa page, offset;
  TC_Entry *tc;
  Octa pAddr;

  if (debugUse) {
    cPrintf("**** MMU: vAddr = #%08X%08X (%s/%s)",
            vAddr.h, vAddr.l, cacheID == ITC ? "ITC" : "DTC",
            accessID == TC_R ? "r" : accessID == TC_W ? "w" : "x");
  }
  page = andOcta(vAddr, currPageMask);
  offset = andOcta(vAddr, currOffsetMask);
  /* do the lookup in any case because of timing */
  switch (cacheID) {
    case ITC:
      tc = assoc(currASN, page, itc, ITC_SIZE);
      break;
    case DTC:
      tc = assoc(currASN, page, dtc, DTC_SIZE);
      break;
    default:
      error("illegal cache ID");
  }
  if ((vAddr.h & SIGN(32)) != 0) {
    /* unmapped address space */
    pAddr.h = vAddr.h & ~SIGN(32);
    pAddr.l = vAddr.l;
  } else {
    /* mapped address space */
    if (tc == NULL) {
      error("TC miss exception not yet handled");
    }
    pAddr = orOcta(tc->frame, offset);
  }
  if (debugUse) {
    cPrintf(", pAddr = #%08X%08X ****\n", pAddr.h, pAddr.l);
  }
  return pAddr;
}


/**************************************************************/


Tetra mmuGetInstr(Octa vAddr) {
  return memoryReadTetra(v2p(vAddr, ITC, TC_X));
}


void mmuPutInstr(Octa vAddr, Tetra instr) {
  memoryWriteTetra(v2p(vAddr, ITC, TC_W), instr);
}


Octa mmuReadOcta(Octa vAddr) {
  return memoryReadOcta(v2p(vAddr, DTC, TC_R));
}


Tetra mmuReadTetra(Octa vAddr) {
  return memoryReadTetra(v2p(vAddr, DTC, TC_R));
}


Wyde mmuReadWyde(Octa vAddr) {
  return memoryReadWyde(v2p(vAddr, DTC, TC_R));
}


Byte mmuReadByte(Octa vAddr) {
  return memoryReadByte(v2p(vAddr, DTC, TC_R));
}


void mmuWriteOcta(Octa vAddr, Octa data) {
  memoryWriteOcta(v2p(vAddr, DTC, TC_W), data);
}


void mmuWriteTetra(Octa vAddr, Tetra data) {
  memoryWriteTetra(v2p(vAddr, DTC, TC_W), data);
}


void mmuWriteWyde(Octa vAddr, Wyde data) {
  memoryWriteWyde(v2p(vAddr, DTC, TC_W), data);
}


void mmuWriteByte(Octa vAddr, Byte data) {
  memoryWriteByte(v2p(vAddr, DTC, TC_W), data);
}


/**************************************************************/


void mmuCtrl(int s, int n) {
  currOffsetMask = incOcta(shlOcta(one, s), -1);
  currPageMask = xorOcta(currOffsetMask, negOne);
  currASN = n;
  if (debugCtrl) {
    cPrintf("**** MMU: currPageMask = #%08X%08X, currASN = %d ****\n",
            currPageMask.h, currPageMask.l, currASN);
  }
}


int mmuUpdate(Octa key) {
  Octa page;
  int asn;
  int prot;
  int whereHit;
  TC_Entry *tc;

  page.h = key.h;
  page.l = key.l & ~0x1FFF;
  asn = (key.l >> 3) & 0x3FF;
  prot = key.l & 0x7;
  whereHit = 0;
  /* check ITC */
  tc = assoc(asn, page, itc, ITC_SIZE);
  if (tc != NULL) {
    /* key is present in ITC */
    if (prot != 0) {
      /* replace protection code */
      tc->prot = prot;
      if (debugCtrl) {
        cPrintf("**** MMU: ITC entry for #%08X%08X/%d, prot = %c%c%c ****\n",
                page.h, page.l, asn,
                prot & 4 ? 'r' : '-',
                prot & 2 ? 'w' : '-',
                prot & 1 ? 'x' : '-');
      }
    } else {
      /* remove key from TC */
      tc->valid = FALSE;
      if (debugCtrl) {
        cPrintf("**** MMU: ITC entry for #%08X%08X/%d removed ****\n",
                page.h, page.l, asn);
      }
    }
    whereHit += 1;
  }
  /* check DTC */
  tc = assoc(asn, page, dtc, DTC_SIZE);
  if (tc != NULL) {
    /* key is present in DTC */
    if (prot != 0) {
      /* replace protection code */
      tc->prot = prot;
      if (debugCtrl) {
        cPrintf("**** MMU: DTC entry for #%08X%08X/%d, prot = %c%c%c ****\n",
                page.h, page.l, asn,
                prot & 4 ? 'r' : '-',
                prot & 2 ? 'w' : '-',
                prot & 1 ? 'x' : '-');
      }
    } else {
      /* remove key from TC */
      tc->valid = FALSE;
      if (debugCtrl) {
        cPrintf("**** MMU: DTC entry for #%08X%08X/%d removed ****\n",
                page.h, page.l, asn);
      }
    }
    whereHit += 2;
  }
  if (debugCtrl && whereHit == 0) {
    cPrintf("**** MMU: no hit for #%08X%08X/%d ****\n",
            page.h, page.l, asn);
  }
  /* return where a hit was found */
  return whereHit;
}


static void showEntry(char *name, int slot, TC_Entry *tc) {
  Octa key;
  Octa xlt;

  cPrintf("%s[%02d]: ", name, slot);
  if (tc->valid) {
    key = incOcta(tc->page, tc->asn << 3);
    xlt = incOcta(shrOcta(tc->frame, 10), tc->prot);
    cPrintf("key = #%08X%08X  -->  xlt = #%02X%08X",
            key.h, key.l, xlt.h, xlt.l);
  } else {
    cPrintf("---- invalid ----");
  }
  cPrintf("\n");
}


void mmuShowAll(int cacheID) {
  char *name;
  TC_Entry *cache;
  int size;
  TC_Entry *tc;
  int i;

  switch (cacheID) {
    case ITC:
      name = "ITC";
      cache = itc;
      size = ITC_SIZE;
      break;
    case DTC:
      name = "DTC";
      cache = dtc;
      size = DTC_SIZE;
      break;
    default:
      error("illegal cache ID");
  }
  tc = cache;
  for (i = 0; i < size; i++) {
    showEntry(name, i, tc);
    tc++;
  }
}


void mmuRemoveAll(int cacheID) {
  char *name;
  TC_Entry *cache;
  int size;
  TC_Entry *tc;
  int i;

  switch (cacheID) {
    case ITC:
      name = "ITC";
      cache = itc;
      size = ITC_SIZE;
      break;
    case DTC:
      name = "DTC";
      cache = dtc;
      size = DTC_SIZE;
      break;
    default:
      error("illegal cache ID");
  }
  tc = cache;
  for (i = 0; i < size; i++) {
    tc->valid = FALSE;
    tc++;
  }
  if (debugCtrl) {
    cPrintf("**** MMU: all entries removed from %s ****\n", name);
  }
}


void mmuShowEntry(int cacheID, Octa key) {
  char *name;
  TC_Entry *cache;
  int size;
  Octa page;
  int asn;
  TC_Entry *tc;

  switch (cacheID) {
    case ITC:
      name = "ITC";
      cache = itc;
      size = ITC_SIZE;
      break;
    case DTC:
      name = "DTC";
      cache = dtc;
      size = DTC_SIZE;
      break;
    default:
      error("illegal cache ID");
  }
  page.h = key.h;
  page.l = key.l & ~0x1FFF;
  asn = (key.l >> 3) & 0x3FF;
  tc = assoc(asn, page, cache, size);
  if (tc == NULL) {
    cPrintf("---- not mapped ----\n");
  } else {
    showEntry(name, tc - cache, tc);
  }
}


void mmuRemoveEntry(int cacheID, Octa key) {
  char *name;
  TC_Entry *cache;
  int size;
  Octa page;
  int asn;
  TC_Entry *tc;

  switch (cacheID) {
    case ITC:
      name = "ITC";
      cache = itc;
      size = ITC_SIZE;
      break;
    case DTC:
      name = "DTC";
      cache = dtc;
      size = DTC_SIZE;
      break;
    default:
      error("illegal cache ID");
  }
  page.h = key.h;
  page.l = key.l & ~0x1FFF;
  asn = (key.l >> 3) & 0x3FF;
  tc = assoc(asn, page, cache, size);
  if (tc != NULL) {
    tc->valid = FALSE;
  }
  if (debugCtrl) {
    cPrintf("**** MMU: key #%08X%08X removed from %s ****\n",
            key.h, key.l, name);
  }
}


void mmuInsertEntry(int cacheID, Octa key, Octa xlt) {
  char *name;
  TC_Entry *cache;
  int size;
  Octa page;
  int asn;
  TC_Entry *tc;
  int slot;

  switch (cacheID) {
    case ITC:
      name = "ITC";
      cache = itc;
      size = ITC_SIZE;
      break;
    case DTC:
      name = "DTC";
      cache = dtc;
      size = DTC_SIZE;
      break;
    default:
      error("illegal cache ID");
  }
  page.h = key.h;
  page.l = key.l & ~0x1FFF;
  asn = (key.l >> 3) & 0x3FF;
  tc = assoc(asn, page, cache, size);
  if (tc == NULL) {
    /* key is not yet mapped, use random replacement strategy */
    slot = rand() % size;
    tc = &cache[slot];
    tc->valid = TRUE;
    tc->page = page;
    tc->asn = asn;
  }
  /* key is (now) mapped */
  tc->frame.h = xlt.h;
  tc->frame.l = xlt.l & ~0x7;
  tc->frame = shlOcta(tc->frame, 10);
  tc->prot = key.l & 0x7;
  if (debugCtrl) {
    cPrintf("**** MMU: key = #%08X%08X  -->  xlt = #%02X%08X in %s ****\n",
            key.h, key.l, xlt.h, xlt.l, name);
  }
}


void mmuReset(void) {
  cPrintf("Resetting MMU...\n");
  mmuRemoveAll(ITC);
  mmuRemoveAll(DTC);
}


void mmuInit(void) {
  mmuReset();
}


void mmuExit(void) {
}
