/*
 * disasm.c -- disassembler
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "arith.h"
#include "spreg.h"
#include "instr.h"
#include "disasm.h"


static char instrBuffer[100];


static void disasmI8I8I8(char *opName, int x, int y, int z) {
  sprintf(instrBuffer, "%-7s #%02X,#%02X,#%02X", opName, x, y, z);
}


static void disasmRRR(char *opName, int x, int y, int z) {
  sprintf(instrBuffer, "%-7s $%d,$%d,$%d", opName, x, y, z);
}


static void disasmRRI8(char *opName, int x, int y, int z) {
  sprintf(instrBuffer, "%-7s $%d,$%d,#%02X", opName, x, y, z);
}


static void disasmRI8R(char *opName, int x, int y, int z) {
  sprintf(instrBuffer, "%-7s $%d,#%02X,$%d", opName, x, y, z);
}


static void disasmRI8I8(char *opName, int x, int y, int z) {
  sprintf(instrBuffer, "%-7s $%d,#%02X,#%02X", opName, x, y, z);
}


static void disasmI8RR(char *opName, int x, int y, int z) {
  sprintf(instrBuffer, "%-7s #%02X,$%d,$%d", opName, x, y, z);
}


static void disasmI8RI8(char *opName, int x, int y, int z) {
  sprintf(instrBuffer, "%-7s #%02X,$%d,#%02X", opName, x, y, z);
}


static void disasmRI16(char *opName, int x, int yz) {
  sprintf(instrBuffer, "%-7s $%d,#%04X", opName, x, yz);
}


static void disasmI8I16(char *opName, int x, int yz) {
  sprintf(instrBuffer, "%-7s #%02X,#%04X", opName, x, yz);
}


static void disasmRF16(char *opName, int x, int yz, octa locus) {
  int offset;
  octa target;

  offset = yz << 2;
  target = incr(locus, offset);
  sprintf(instrBuffer, "%-7s $%d,#%08X%08X", opName, x, target.h, target.l);
}


static void disasmRB16(char *opName, int x, int yz, octa locus) {
  int offset;
  octa target;

  offset = (yz - 0x10000) << 2;
  target = incr(locus, offset);
  sprintf(instrBuffer, "%-7s $%d,#%08X%08X", opName, x, target.h, target.l);
}


static void disasmF24(char *opName, int xyz, octa locus) {
  int offset;
  octa target;

  offset = xyz << 2;
  target = incr(locus, offset);
  sprintf(instrBuffer, "%-7s #%08X%08X", opName, target.h, target.l);
}


static void disasmB24(char *opName, int xyz, octa locus) {
  int offset;
  octa target;

  offset = (xyz - 0x1000000) << 2;
  target = incr(locus, offset);
  sprintf(instrBuffer, "%-7s #%08X%08X", opName, target.h, target.l);
}


static void disasmRZ8S8(char *opName, int x, int z) {
  if (z >= 0 && z <= 31) {
    /* this is the normal case */
    sprintf(instrBuffer, "%-7s $%d,%s", opName, x, spRegCodeTbl[z]->name);
  } else {
    /* this is illegal */
    sprintf(instrBuffer, "%-7s $%d,#%02X", opName, x, z);
  }
}


static void disasmS8Z8R(char *opName, int x, int z) {
  if (x >= 0 && x <= 31) {
    /* this is the normal case */
    sprintf(instrBuffer, "%-7s %s,$%d", opName, spRegCodeTbl[x]->name, z);
  } else {
    /* this is illegal */
    sprintf(instrBuffer, "%-7s #%02X,$%d", opName, x, z);
  }
}


static void disasmS8Z8I8(char *opName, int x, int z) {
  if (x >= 0 && x <= 31) {
    /* this is the normal case */
    sprintf(instrBuffer, "%-7s %s,#%02X", opName, spRegCodeTbl[x]->name, z);
  } else {
    /* this is illegal */
    sprintf(instrBuffer, "%-7s #%02X,#%02X", opName, x, z);
  }
}


static void disasmRZ8R(char *opName, int x, int z) {
  sprintf(instrBuffer, "%-7s $%d,$%d", opName, x, z);
}


static void disasmRZ8I8(char *opName, int x, int z) {
  sprintf(instrBuffer, "%-7s $%d,#%02X", opName, x, z);
}


static void disasmRZ8Z8(char *opName, int x) {
  sprintf(instrBuffer, "%-7s $%d,0", opName, x);
}


static void disasmZ8Z8R(char *opName, int z) {
  sprintf(instrBuffer, "%-7s 0,$%d", opName, z);
}


static void disasmZ8Z8I8(char *opName, int z) {
  sprintf(instrBuffer, "%-7s #%02X", opName, z);
}


char *disasm(tetra instr, octa locus) {
  Instr *ip;

  ip = instrCodeTbl[OPCODE(instr)];
  if (ip == NULL) {
    disasmI8I8I8("???", DST(instr), SRC1(instr), SRC2(instr));
  } else {
    switch (ip->format) {
      case FMT_I8I8I8:
        disasmI8I8I8(ip->name, DST(instr), SRC1(instr), SRC2(instr));
        break;
      case FMT_RRR:
        disasmRRR(ip->name, DST(instr), SRC1(instr), SRC2(instr));
        break;
      case FMT_RRI8:
        disasmRRI8(ip->name, DST(instr), SRC1(instr), SRC2(instr));
        break;
      case FMT_RI8R:
        disasmRI8R(ip->name, DST(instr), SRC1(instr), SRC2(instr));
        break;
      case FMT_RI8I8:
        disasmRI8I8(ip->name, DST(instr), SRC1(instr), SRC2(instr));
        break;
      case FMT_I8RR:
        disasmI8RR(ip->name, DST(instr), SRC1(instr), SRC2(instr));
        break;
      case FMT_I8RI8:
        disasmI8RI8(ip->name, DST(instr), SRC1(instr), SRC2(instr));
        break;
      case FMT_RI16:
        disasmRI16(ip->name, DST(instr), LIT16(instr));
        break;
      case FMT_I8I16:
        disasmI8I16(ip->name, DST(instr), LIT16(instr));
        break;
      case FMT_RF16:
        disasmRF16(ip->name, DST(instr), LIT16(instr), locus);
        break;
      case FMT_RB16:
        disasmRB16(ip->name, DST(instr), LIT16(instr), locus);
        break;
      case FMT_F24:
        disasmF24(ip->name, LIT24(instr), locus);
        break;
      case FMT_B24:
        disasmB24(ip->name, LIT24(instr), locus);
        break;
      case FMT_RZ8S8:
        disasmRZ8S8(ip->name, DST(instr), SRC2(instr));
        break;
      case FMT_S8Z8R:
        disasmS8Z8R(ip->name, DST(instr), SRC2(instr));
        break;
      case FMT_S8Z8I8:
        disasmS8Z8I8(ip->name, DST(instr), SRC2(instr));
        break;
      case FMT_RZ8R:
        disasmRZ8R(ip->name, DST(instr), SRC2(instr));
        break;
      case FMT_RZ8I8:
        disasmRZ8I8(ip->name, DST(instr), SRC2(instr));
        break;
      case FMT_RZ8Z8:
        disasmRZ8Z8(ip->name, DST(instr));
        break;
      case FMT_Z8Z8R:
        disasmZ8Z8R(ip->name, SRC2(instr));
        break;
      case FMT_Z8Z8I8:
        disasmZ8Z8I8(ip->name, SRC2(instr));
        break;
      default:
        error("unknown instruction format");
    }
  }
  return instrBuffer;
}
