/*
 * vat.c -- virtual address translation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "arith.h"
#include "vat.h"
#include "spreg.h"
#include "cpu.h"
#include "memory.h"


static bool debugNeg = false;
static bool debugPos = true;

static octa rv;
static unsigned int b[5];
static unsigned int size;
static unsigned int root;
static unsigned int asn;
static unsigned int fnc;


static void getTranslationValues(void) {
  rv = cpuGetGR(rV);
  b[0] = 0;
  b[1] = (rv.h >> 28) & 0xF;
  b[2] = (rv.h >> 24) & 0xF;
  b[3] = (rv.h >> 20) & 0xF;
  b[4] = (rv.h >> 16) & 0xF;
  size = (rv.h >> 8) & 0xFF;
  root = ((rv.h & 0xFF) << 19) | ((rv.l >> 13) & 0x7FFFF);
  asn = (rv.l >> 3) & 0x3FF;
  fnc = rv.l & 0x7;
}


static void printOcta(octa val) {
  cPrintf("#%08X%08X", val.h, val.l);
}


static bool vatLoad(octa addr, octa *dataPtr) {
  octa data;

  if (addr.h != MEM_BASE_H || addr.l >= memoryGetSize()) {
    return false;
  }
  data = memoryRead(addr);
  if (debugPos) {
    cPrintf("VAT: load from ");
    printOcta(addr);
    cPrintf(" yields ");
    printOcta(data);
    cPrintf("\n");
  }
  *dataPtr = data;
  return true;
}


char *vat(octa virt, octa *physPtr) {
  octa phys;
  int segment;
  octa mask;
  octa page;
  octa offset;
  int level, i;
  int a[5];
  octa base;
  octa address;
  octa descr;

  if (virt.h & SIGN(32)) {
    if (debugNeg) {
      cPrintf("VAT: virtual address = ");
      printOcta(virt);
      cPrintf("\n");
    }
    phys.h = virt.h & ~SIGN(32);
    phys.l = virt.l;
    if (debugNeg) {
      cPrintf("VAT: physical address = ");
      printOcta(phys);
      cPrintf("\n");
    }
  } else {
    if (debugPos) {
      cPrintf("VAT: virtual address = ");
      printOcta(virt);
      cPrintf("\n");
    }
    getTranslationValues();
    if (size < 13) {
      return "page size is less than 2^13";
    }
    if (size > 48) {
      return "page size is greater than 2^48";
    }
    segment = (virt.h >> 29) & 0x3;
    if (debugPos) {
      cPrintf("VAT: segment number = %d\n", segment);
    }
    virt.h &= 0x1FFFFFFF;
    if (debugPos) {
      cPrintf("VAT: segment-relative virtual address = ");
      printOcta(virt);
      cPrintf("\n");
    }
    mask = incr(shift_left(one_octa, size), -1);
    if (debugPos) {
      cPrintf("VAT: page mask = ");
      printOcta(mask);
      cPrintf("\n");
    }
    page = shift_right(virt, size, 1);
    offset = oand(virt, mask);
    if (debugPos) {
      cPrintf("VAT: page = ");
      printOcta(page);
      cPrintf(", offset = ");
      printOcta(offset);
      cPrintf("\n");
    }
    level = 0;
    for (i = 0; i < 5; i++) {
      a[i] = shift_right(page, 10 * i, 1).l & 0x3FF;
      if (a[i] != 0) {
        level = i;
      }
    }
    if (debugPos) {
      cPrintf("VAT: page number in radix 1024 = ");
      cPrintf("(#%03X #%03X #%03X #%03X #%03X)\n",
              a[4], a[3], a[2], a[1], a[0]);
      cPrintf("VAT: translation chain = VA --> ");
      if (level >= 4) {
        cPrintf("PTP(4) --> ");
      }
      if (level >= 3) {
        cPrintf("PTP(3) --> ");
      }
      if (level >= 2) {
        cPrintf("PTP(2) --> ");
      }
      if (level >= 1) {
        cPrintf("PTP(1) --> ");
      }
      cPrintf("PTE\n");
    }
    /* do first stage of lookup */
    base.h = 0;
    base.l = root;
    base = incr(base, b[segment] + level);
    base = shift_left(base, 13);
    address = incr(base, a[level] << 3);
    if (debugPos) {
      cPrintf("VAT: looking up memory at address\n");
      cPrintf("     ((#%08X + #%02X + %d) << 13) + (#%03X << 3) = ",
              root, b[segment], level, a[level]);
      printOcta(address);
      cPrintf("\n");
    }
    if (!vatLoad(address, &descr)) {
      return "cannot load PTP/PTE";
    }
    /* interpret auxiliary PTP values */
    for (; level > 0; level--) {
      if (debugPos) {
        cPrintf("VAT: PTP value at level %d = ", level);
        printOcta(descr);
        cPrintf("\n");
      }
      mask = incr(shift_left(one_octa, 13), -1);
      mask.h |= 0x80000000;
      mask = oxor(mask, neg_one);
      base = oand(descr, mask);
      address = incr(base, a[level - 1] << 3);
      if (debugPos) {
        cPrintf("VAT: looking up memory at address\n");
        cPrintf("     ");
        printOcta(base);
        cPrintf(" + (#%03X << 3) = ", a[level - 1]);
        printOcta(address);
        cPrintf("\n");
      }
      if (!vatLoad(address, &descr)) {
        return "cannot load PTP/PTE";
      }
    }
    /* interpret PTE value */
    if (debugPos) {
      cPrintf("VAT: PTE value = ");
      printOcta(descr);
      cPrintf("\n");
    }
    mask = incr(shift_left(one_octa, size), -1);
    mask.h |= 0xFFFF0000;
    mask = oxor(mask, neg_one);
    base = oand(descr, mask);
    phys = oor(base, offset);
    if (debugPos) {
      cPrintf("VAT: physical page base address = ");
      printOcta(base);
      cPrintf("\n");
      cPrintf("VAT: physical address = ");
      printOcta(phys);
      cPrintf("\n");
    }
  }
  *physPtr = phys;
  return NULL;
}
