/*
 * ld.c -- load a GIMMIX object file, i.e., convert it to mmx format
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>


/**************************************************************/


#define MM		0x98		/* escape code of mmo format */
#define LOP_QUOTE	0x00		/* quotation lopcode */
#define LOP_LOC		0x01		/* location lopcode */
#define LOP_SKIP	0x02		/* skip lopcode */
#define LOP_FIXO	0x03		/* octabyte-fix lopcode */
#define LOP_FIXR	0x04		/* relative-fix lopcode */
#define LOP_FIXRX	0x05		/* extended relative-fix lopcode */
#define LOP_FILE	0x06		/* file name lopcode */
#define LOP_LINE	0x07		/* file position lopcode */
#define LOP_SPEC	0x08		/* special hook lopcode */
#define LOP_PRE		0x09		/* preamble lopcode */
#define LOP_POST	0x0A		/* postamble lopcode */
#define LOP_STAB	0x0B		/* symbol table lopcode */
#define LOP_END		0x0C		/* end-it-all lopcode */


#define B0(t)		((t >> 24) & 0xFF)
#define B1(t)		((t >> 16) & 0xFF)
#define B2(t)		((t >>  8) & 0xFF)
#define B3(t)		((t >>  0) & 0xFF)

#define W0(t)		((t >> 16) & 0xFFFF)
#define W1(t)		((t >>  0) & 0xFFFF)


/**************************************************************/


typedef unsigned char U8;
typedef signed char S8;
typedef unsigned short U16;
typedef signed short S16;
typedef unsigned int U32;
typedef signed int S32;


typedef U8 byte;
typedef U16 wyde;
typedef U32 tetra;
typedef struct {
  tetra h;
  tetra l;
} octa;


typedef enum { false, true } bool;


/**************************************************************/


bool debugLoad = false;
bool debugDump = false;

char *mmoName;			/* input file name */
FILE *mmoFile;			/* input file */
char *mmxName;			/* output file name */
FILE *mmxFile;			/* output file */

octa currLoc;			/* current location */


/**************************************************************/


void error(char *fmt, ...) {
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "Error: ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
  exit(1);
}


/**************************************************************/


octa incOcta(octa y, int delta) {
  octa x;

  x.h = y.h;
  x.l = y.l + delta;
  if (delta >= 0) {
    if (x.l < y.l) {
      x.h++;
    }
  } else {
    if (x.l > y.l) {
      x.h--;
    }
  }
  return x;
}


static bool hasUnreadTetra = false;
static tetra unreadValue;


tetra readTetra(void) {
  byte b[4];
  tetra value;

  if (hasUnreadTetra) {
    hasUnreadTetra = false;
    return unreadValue;
  }
  if (fread(b, sizeof(byte), 4, mmoFile) != 4) {
    error("unexpected end of file '%s'", mmoName);
  }
  value = ((tetra) b[0]) << 24 |
          ((tetra) b[1]) << 16 |
          ((tetra) b[2]) <<  8 |
          ((tetra) b[3]) <<  0;
  return value;
}


void unreadTetra(tetra value) {
  hasUnreadTetra = true;
  unreadValue = value;
}


/**************************************************************/


typedef struct {
  tetra tetra;			/* the simulated tetrabyte */
  bool written;			/* true iff tetra was written to */
} MemTetra;


typedef struct memNode {
  octa loc;			/* base address of 512 tetrabytes */
  tetra stamp;			/* time stamp for treap balancing */
  struct memNode *left;		/* left subtree of binary search tree */
  struct memNode *right;	/* right subtree of binary search tree */
  MemTetra data[512];		/* the chunk of simulated tetrabytes */
} MemNode;


static tetra priority = 314159265;
static MemNode *memRoot = NULL;
static MemNode *memLast = NULL;


static MemNode *memNew(void) {
  MemNode *p;

  p = calloc(1, sizeof(MemNode));
  if (p == NULL) {
    error("cannot allocate memory");
  }
  p->stamp = priority;
  priority += 0x9E3779B9;
  return p;
}


MemTetra *memFind(octa addr) {
  octa key;
  int offset;
  MemNode *p;
  MemNode **q;
  MemNode **l, **r;

  key.h = addr.h;
  key.l = addr.l & 0xFFFFF800;
  offset = addr.l & 0x7FC;
  p = memLast;
  if (p == NULL || p->loc.l != key.l || p->loc.h != key.h) {
    p = memRoot;
    while (p != NULL) {
      if (key.l == p->loc.l && key.h == p->loc.h) {
        memLast = p;
        return &p->data[offset >> 2];
      }
      if ((key.l < p->loc.l && key.h <= p->loc.h) || key.h < p->loc.h) {
        p = p->left;
      } else {
        p = p->right;
      }
    }
    p = memRoot;
    q = &memRoot;
    while (p != NULL && p->stamp < priority) {
      if ((key.l < p->loc.l && key.h <= p->loc.h) || key.h < p->loc.h) {
        q = &p->left;
      } else {
        q = &p->right;
      }
      p = *q;
    }
    *q = memNew();
    (*q)->loc = key;
    l = &(*q)->left;
    r = &(*q)->right;
    while (p != NULL) {
      if ((key.l < p->loc.l && key.h <= p->loc.h) || key.h < p->loc.h) {
        *r = p;
        r = &p->left;
        p = *r;
      } else {
        *l = p;
        l = &p->right;
        p = *l;
      }
    }
    *l = NULL;
    *r = NULL;
    p = *q;
    memLast = p;
  }
  return &p->data[offset >> 2];
}


/**************************************************************/


static octa currentDumpLoc = {
  -1, -1
};


void dumpOcta(octa addr, octa data) {
  if (debugDump) {
    printf("DUMP #%08X%08X: #%08X%08X\n", addr.h, addr.l, data.h, data.l);
  }
  if ((addr.h & 0x80000000) == 0) {
    error("non-physical address encountered, cannot load");
  }
  if ((addr.h & 0x7FFF0000) != 0) {
    error("address is outside of physical memory address range");
  }
  addr.h &= 0x0000FFFF;
  if (addr.h != currentDumpLoc.h || addr.l != currentDumpLoc.l) {
    fprintf(mmxFile, "%04X%08X:\n", addr.h, addr.l);
    currentDumpLoc.h = addr.h;
    currentDumpLoc.l = addr.l;
  }
  fprintf(mmxFile, " %08X%08X\n", data.h, data.l);
  currentDumpLoc = incOcta(currentDumpLoc, 8);
}


void dump(MemNode *p) {
  int i;
  octa addr, data;

  if (p->left != NULL) {
    dump(p->left);
  }
  addr = p->loc;
  for (i = 0; i < 512; i += 2) {
    if (p->data[i + 0].written || p->data[i + 1].written) {
      data.h = p->data[i + 0].tetra;
      data.l = p->data[i + 1].tetra;
      dumpOcta(addr, data);
    }
    addr = incOcta(addr, 8);
  }
  if (p->right != NULL) {
    dump(p->right);
  }
}


/**************************************************************/


void usage(char *myself) {
  fprintf(stderr, "Usage: %s <object file> <hex file>\n", myself);
  exit(1);
}


void loadTetra(octa loc, tetra val) {
  MemTetra *ll;

  if (debugLoad) {
    printf("LOAD #%08X%08X: #%08X\n", loc.h, loc.l, val);
  }
  ll = memFind(loc);
  ll->tetra ^= val;
  ll->written = true;
}


int main(int argc, char *argv[]) {
  int i;
  tetra t;
  bool postambleSeen;
  tetra delta;
  tetra increment;
  octa temp;

  /* process command line */
  if (argc != 3) {
    usage(argv[0]);
  }
  mmoName = argv[1];
  mmxName = argv[2];
  /* initialize everything */
  mmoFile = fopen(mmoName, "rb");
  if (mmoFile == NULL) {
    error("cannot open object file '%s'", mmoName);
  }
  mmxFile = fopen(mmxName, "wb");
  if (mmxFile == NULL) {
    error("cannot open hex file '%s'", mmxName);
  }
  currLoc.h = 0;
  currLoc.l = 0;
  /* load the preamble */
  t = readTetra();
  if (B0(t) != MM || B1(t) != LOP_PRE) {
    error("input file '%s' is not a GIMMIX object file", mmoName);
  }
  if (B2(t) != 1) {
    error("object format version %d not supported", B2(t));
  }
  if (B3(t) > 0) {
    i = B3(t);
    t = readTetra();
    for (i--; i > 0; i--) {
      t = readTetra();
    }
  }
  /* load items */
  postambleSeen = false;
  while (!postambleSeen) {
    t = readTetra();
    if (B0(t) != MM) {
      /* tetra is a normal item */
      loadTetra(currLoc, t);
      currLoc = incOcta(currLoc, 4);
      currLoc.l &= -4;
    } else {
      /* tetra is a loader command */
      switch (B1(t)) {
        case LOP_QUOTE:
          if (W1(t) != 1) {
            error("YZ field of lop_quote must be 1");
          }
          t = readTetra();
          loadTetra(currLoc, t);
          currLoc = incOcta(currLoc, 4);
          currLoc.l &= -4;
          break;
        case LOP_LOC:
          if (B3(t) == 2) {
            currLoc.h = (B2(t) << 24) + readTetra();
          } else
          if (B3(t) == 1) {
            currLoc.h = B2(t) << 24;
          } else {
            error("Z field of lop_loc must be 1 or 2");
          }
          currLoc.l = readTetra();
          break;
        case LOP_SKIP:
          delta = W1(t);
          currLoc = incOcta(currLoc, delta);
          break;
        case LOP_FIXO:
          if (B3(t) == 2) {
            temp.h = (B2(t) << 24) + readTetra();
          } else
          if (B3(t) == 1) {
            temp.h = B2(t) << 24;
          } else {
            error("Z field of lop_fixo must be 1 or 2");
          }
          temp.l = readTetra();
          loadTetra(temp, currLoc.h);
          loadTetra(incOcta(temp, 4), currLoc.l);
          break;
        case LOP_FIXR:
          delta = W1(t);
          increment = -delta << 2;
          temp = incOcta(currLoc, increment);
          loadTetra(temp, delta);
          break;
        case LOP_FIXRX:
          if (B2(t) != 0) {
            error("Y field of lop_fixrx must be 0");
          }
          if (B3(t) != 16 && B3(t) != 24) {
            error("Z field of lop_fixrx must be 16 or 24");
          }
          delta = readTetra();
          if ((delta & 0xFE000000) != 0) {
            error("increment of lop_fixrx is too large");
          }
          if (delta >= 0x1000000) {
            increment = -((delta & 0xFFFFFF) - (1 << B3(t))) << 2;
          } else {
            increment = -delta << 2;
          }
          temp = incOcta(currLoc, increment);
          loadTetra(temp, delta);
          break;
        case LOP_FILE:
          if (B3(t) != 0) {
            for (i = B3(t); i > 0; i--) {
              t = readTetra();
            }
          }
          break;
        case LOP_LINE:
          break;
        case LOP_SPEC:
          while (1) {
            t = readTetra();
            if (B0(t) == MM) {
              if (B1(t) != LOP_QUOTE || W1(t) != 1) {
                /* end of special data */
                unreadTetra(t);
                break;
              }
              t = readTetra();
            }
          }
          break;
        case LOP_PRE:
          error("cannot have another preamble");
          break;
        case LOP_POST:
          postambleSeen = true;
          if (B2(t) != 0) {
            error("Y field of lop_post must be zero");
          }
          if (B3(t) < 32) {
            error("Z field of lop_post is less than 32");
          }
          break;
        case LOP_STAB:
          error("symbol table must follow postamble");
          break;
        case LOP_END:
          error("symbol table cannot end before it begins");
          break;
        default:
          error("unknown loader command #%02X", B1(t));
          break;
      }
    }
  }
  /* load the postamble */
  if (B3(t) != 255) {
    error("global register pre-loading not allowed");
  }
  temp.h = readTetra();
  temp.l = readTetra();
  if ((temp.h & 0x80000000) == 0) {
    error("symbol 'Main' has non-physical address");
  }
  if ((temp.h & 0x7FFF0000) != 0) {
    error("symbol 'Main' is outside of physical memory address range");
  }
  /* dump the loaded program */
  dump(memRoot);
  /* close object file and return */
  fclose(mmoFile);
  return 0;
}
